class EarlyStopping:
    def __init__(self, delta = -1e-3, patience = 10, skip = 0):
        """
        Args:
            delta (float): the smallest change that considered as an improvement.
                If is positive, means larger is better, negative means smaller is better.
            patience (int): how many times will be stop when score has no improvement
            skip (int): how many rounds should skip after start training
        """
        self.direction = 1.0 if delta > 0 else -1.0
        self.delta = delta * self.direction
        self.patience = patience
        self.skip = skip
        
        self.reset()

    
    def get_best_state(self):
        """get best state of model
        """
        return self.best_state
    

    def reset(self):
        """
        """
        self.best_score = float('inf') * (-self.direction)
        self.best_state = None
        self._times = 0
        self._round = -1
    

    def __call__(self, model, *args, **kwargs):
        self._round += 1

        # set skip round
        if self._round < self.skip:
            return False
        
        score = self.scoring(model, *args, **kwargs)
        diff = (score - self.best_score) * self.direction
        
        if diff > self.delta:
            self.best_state = model.state_dict()
            self.best_score = score
            self._times = 0
            return False
        
        self._times += 1
        if self._times >= self.patience:
            # model.load_state_dict(self.best_state)
            return True
    
    def scoring(self, model, history, epoch = None):
        """scoring function
        """
        return history['loss'].mean()



import torch
from torch.nn import Module

from .functional import focal_loss


class FocalLoss(Module):
    def __init__(self, alpha = 1., gamma = 2., reduction = 'mean'):
        super(FocalLoss, self).__init__()
        
        self.alpha = alpha
        self.gamma = gamma
        self.reduction = reduction

    def forward(self, input, target):
        return focal_loss(
            input,
            target,
            alpha = self.alpha,
            gamma = self.gamma,
            reduction = self.reduction,
        )
        




from time import sleep
from .progress import Progress


def test_progress():
    for i in Progress(range(100)):
        sleep(0.01)
        pass


def test_progress_size():
    p = Progress(range(9527))
    assert p.size == 9527


from joblib import Parallel, delayed

import numpy as np
import pandas as pd

from scipy import stats
from .metrics import KS, AUC, F1, PSI

from .utils import (
    np_count,
    np_unique,
    to_ndarray,
    feature_splits,
    is_continuous,
    inter_feature,
    split_target,
)

from .utils.decorator import Decorator, support_dataframe

STATS_EMPTY = np.nan

def gini(target):
    """get gini index of a feature

    Args:
        target (array-like): list of target that will be calculate gini

    Returns:
        number: gini value
    """
    target = to_ndarray(target)
    v, c = np.unique(target, return_counts = True)

    return 1 - ((c / target.size) ** 2).sum()

def _gini_cond(feature, target):
    """private conditional gini function

    Args:
        feature (numpy.ndarray)
        target (numpy.ndarray)

    Returns:
        number: conditional gini value
    """
    size = feature.size

    value = 0
    for v, c in zip(*np_unique(feature, return_counts = True)):
        target_series = target[feature == v]
        value += c / size * gini(target_series)

    return value

@support_dataframe
def gini_cond(feature, target):
    """get conditional gini index of a feature

    Args:
        feature (array-like)
        target (array-like)

    Returns:
        number: conditional gini value. If feature is continuous, it will return the best gini value when the feature bins into two groups
    """
    if not is_continuous(feature):
        return _gini_cond(feature, target)

    # find best split for continuous data
    splits = feature_splits(feature, target)
    best = 999

    for f in inter_feature(feature, splits):
        v = _gini_cond(f, target)
        if v < best:
            best = v
    return best

def entropy(target):
    """get infomation entropy of a feature

    Args:
        target (array-like)

    Returns:
        number: information entropy
    """
    target = to_ndarray(target)
    uni, counts = np.unique(target, return_counts = True)
    prob = counts / len(target)
    entropy = stats.entropy(prob)
    return entropy

def _entropy_cond(feature, target):
    """private conditional entropy func

    Args:
        feature (numpy.ndarray)
        target (numpy.ndarray)

    Returns:
        number: conditional information entropy
    """
    size = len(feature)

    value = 0
    for v, c in zip(*np_unique(feature, return_counts = True)):
        target_series = target[feature == v]
        value += c/size * entropy(target_series)

    return value

@support_dataframe
def entropy_cond(feature, target):
    """get conditional entropy of a feature

    Args:
        feature (array-like)
        target (array-like)

    Returns:
        number: conditional information entropy. If feature is continuous, it will return the best entropy when the feature bins into two groups
    """
    feature = to_ndarray(feature)
    target = to_ndarray(target)

    if not is_continuous(feature):
        return _entropy_cond(feature, target)

    # find best split for continuous data
    splits = feature_splits(feature, target)
    best = 0
    for f in inter_feature(feature, splits):
        v = _entropy_cond(f, target)
        if v > best:
            best = v
    return best


def probability(target, mask = None):
    """get probability of target by mask
    """
    if mask is None:
        return 1, 1

    counts_0 = np_count(target, 0, default = 1)
    counts_1 = np_count(target, 1, default = 1)

    sub_target = target[mask]

    sub_0 = np_count(sub_target, 0, default = 1)
    sub_1 = np_count(sub_target, 1, default = 1)

    y_prob = sub_1 / counts_1
    n_prob = sub_0 / counts_0

    return y_prob, n_prob


def WOE(y_prob, n_prob):
    """get WOE of a group

    Args:
        y_prob: the probability of grouped y in total y
        n_prob: the probability of grouped n in total n

    Returns:
        number: woe value
    """
    return np.log(y_prob / n_prob)


def _IV(feature, target):
    """private information value func

    Args:
        feature (array-like)
        target (array-like)

    Returns:
        number: IV
        Series: IV of each groups
    """
    feature = to_ndarray(feature)
    target = to_ndarray(target)

    iv = {}

    for v in np.unique(feature):
        y_prob, n_prob = probability(target, mask = (feature == v))

        iv[v] = (y_prob - n_prob) * WOE(y_prob, n_prob)

    iv = pd.Series(iv)
    return iv.sum(), iv


@support_dataframe
def IV(feature, target, return_sub = False, **kwargs):
    """get the IV of a feature

    Args:
        feature (array-like)
        target (array-like)
        return_sub (bool): if need return IV of each groups
        n_bins (int): n groups that the feature will bin into
        method (str): the strategy to be used to merge feature, default is 'dt'
        **kwargs (): other options for merge function
    """
    if is_continuous(feature):
        from .merge import merge
        feature = merge(feature, target, **kwargs)

    iv, sub = _IV(feature, target)

    if return_sub:
        return iv, sub
    
    return iv


def badrate(target):
    """calculate badrate

    Args:
        target (array-like): target array which `1` is bad

    Returns:
        float
    """
    return np.sum(target) / len(target)


def VIF(frame):
    """calculate vif

    Args:
        frame (ndarray|DataFrame)

    Returns:
        Series
    """
    index = None
    if isinstance(frame, pd.DataFrame):
        index = frame.columns
        frame = frame.values
    
    from sklearn.linear_model import LinearRegression

    model = LinearRegression(fit_intercept = False)

    l = frame.shape[1]
    vif = np.zeros(l)

    for i in range(l):
        X = frame[:, np.arange(l) != i]
        y = frame[:, i]
        model.fit(X, y)

        pre_y = model.predict(X)

        vif[i] = np.sum(y ** 2) / np.sum((pre_y - y) ** 2)
    
    return pd.Series(vif, index = index)



class indicator(Decorator):
    """indicator decorator
    """
    # indicator name
    name = 'indicator'
    need_merge = False
    dtype = None

    def wrapper(self, *args, **kwargs):
        return self.call(*args, **kwargs)

# default indicators
INDICATORS = {
    'iv': indicator(name = 'iv', need_merge = True)(IV),
    'gini': indicator(name = 'gini')(gini_cond),
    'entropy': indicator(name = 'entropy')(entropy_cond),
    'auc': indicator(name = 'auc', dtype = np.number)(AUC),
    'ks': indicator(name = 'ks', dtype = np.number)(KS),
    'unique': indicator(name = 'unique')(lambda x, *arg: len(np_unique(x))),
}


def column_quality(feature, target, name = 'feature', indicators = [], need_merge = False, **kwargs):
    """calculate quality of a feature

    Args:
        feature (array-like)
        target (array-like)
        name (str): feature's name that will be setted in the returned Series
        indicators (list): list of indicator functions
        need_merge (bool): if need merge feature

    Returns:
        Series: a list of quality with the feature's name
    """
    feature = to_ndarray(feature)
    target = to_ndarray(target)

    if not np.issubdtype(feature.dtype, np.number):
        feature = feature.astype(str)

    # get bin feature
    bin_feature = feature
    if need_merge and is_continuous(feature):
        from .merge import merge
        bin_feature = merge(feature, target, **kwargs)

    res = {}  
    for func in indicators:
        # filter by dtype
        if func.dtype is not None and not isinstance(feature.dtype, func.dtype):
            res[func.name] = STATS_EMPTY
            continue

        # if function need use bin feature
        if func.need_merge:
            res[func.name] = func(bin_feature, target)
            continue
        
        res[func.name] = func(feature, target)

    row = pd.Series(res)

    row.name = name
    return row


def quality(dataframe, target = 'target', cpu_cores = 0, iv_only = False, indicators = ['iv', 'gini', 'entropy', 'unique'], **kwargs):
    """get quality of features in data

    Args:
        dataframe (DataFrame): dataframe that will be calculate quality
        target (str): the target's name in dataframe
        iv_only (bool): `deprecated`. if only calculate IV
        cpu_cores (int): the maximun number of CPU cores will be used, `0` means all CPUs will be used, 
            `-1` means all CPUs but one will be used.

    Returns:
        DataFrame: quality of features with the features' name as row name
    """
    frame, target = split_target(dataframe, target)

    if iv_only:
        import warnings
        warnings.warn(
            """`iv_only` will be deprecated soon,
                please use `indicators = ['iv']` instead!
            """,
            DeprecationWarning,
        )

        dummy_func = lambda x, t: STATS_EMPTY

        indicators = [
            'iv',
            indicator(name = 'gini')(dummy_func),
            indicator(name = 'entropy')(dummy_func),
            'unique',
        ]
    
    
    need_merge = False
    for i, f in enumerate(indicators):
        # replace str type indicator to function
        if isinstance(f, str):
            assert f in INDICATORS

            indicators[i] = INDICATORS[f]
        
        # update need merge flag
        need_merge |= indicators[i].need_merge
    
    
    if cpu_cores < 1:
        cpu_cores = cpu_cores - 1
    
    
    pool = Parallel(n_jobs = cpu_cores)

    jobs = []
    for name, series in frame.iteritems():
        jobs.append(delayed(column_quality)(
            series,
            target,
            name = name,
            indicators = indicators,
            need_merge = need_merge,
            **kwargs
        ))

    rows = pool(jobs)


    return pd.DataFrame(rows).sort_values(
        by = indicators[0].name,
        ascending = False,
    )



"""
toad command line application
"""
import argparse
from .commands import get_plugins


def add_sub(parsers, config):
    """add sub parser by config
    """
    info = config.get('info', {})
    args = config.get('args', [])
    defaults = config.get('defaults', None)

    sub_parser = parsers.add_parser(**info)

    for detail in args:
        flag = detail.pop('flag')
        sub_parser.add_argument(*flag, **detail)

    if defaults:
        sub_parser.set_defaults(**defaults)


def get_parser():
    """get parser
    """
    parser = argparse.ArgumentParser(
        prog = 'toad',
        description = 'Detect data from a csv file',
    )

    subparsers = parser.add_subparsers()

    plugins = get_plugins()
    for plug in plugins:
        add_sub(subparsers, plug.ARGS)

    return parser


def main():
    """
    """
    parser = get_parser()

    args = parser.parse_args()
    if hasattr(args, 'func'):
        args.func(args)


if __name__ == '__main__':
    main()



import numpy as np
import pandas as pd
from .func import save_json, read_json
from functools import wraps, WRAPPER_ASSIGNMENTS



class Decorator:
    """base decorater class
    """
    _fn = None
    _cls = None
    is_class = False

    def __init__(self, *args, is_class = False, **kwargs):
        self.is_class = is_class

        if len(args) == 1 and callable(args[0]):
            self._fn = args[0]
        else:
            self.setup(*args, **kwargs)

    def __call__(self, *args, **kwargs):
        if self._fn is None:
            self._fn = args[0]
            return self

        if self.is_class:
            self._cls = args[0]
            args = args[1:]

        return self.wrapper(*args, **kwargs)


    def __get__(self, instance, type = None):
        self.is_class = True
        self._cls = instance

        @wraps(self._fn)
        def func(*args, **kwargs):
            return self.__call__(instance, *args, **kwargs)

        return func


    def __getattribute__(self, name):
        if name in WRAPPER_ASSIGNMENTS:
            return getattr(self._fn, name)

        return object.__getattribute__(self, name)


    def setup(self, *args, **kwargs):
        for key in kwargs:
            setattr(self, key, kwargs[key])

    def call(self, *args, **kwargs):
        if self._cls is not None:
            args = (self._cls, *args)

        return self._fn(*args, **kwargs)

    def wrapper(self, *args, **kwargs):
        return self.call(*args, **kwargs)


class frame_exclude(Decorator):
    """decorator for exclude columns
    """

    def wrapper(self, X, *args, exclude = None, **kwargs):
        if exclude is not None and isinstance(X, pd.DataFrame):
            X = X.drop(columns = exclude)

        return self.call(X, *args, **kwargs)


class select_dtypes(Decorator):
    """ decorator for select frame by dtypes
    """

    def wrapper(self, X, *args, select_dtypes = None, **kwargs):
        if select_dtypes is not None and isinstance(X, pd.DataFrame):
            X = X.select_dtypes(include = select_dtypes)

        return self.call(X, *args, **kwargs)


class save_to_json(Decorator):
    """support save result to json file
    """
    def wrapper(self, *args, to_json = None, **kwargs):
        res = self.call(*args, **kwargs)

        if to_json is not None:
            save_json(res, to_json)

        return res


class load_from_json(Decorator):
    """support load data from json file
    """
    require_first = False

    def wrapper(self, *args, from_json = None, **kwargs):
        if from_json is not None:
            obj = read_json(from_json)
            args = (obj, *args)
        
        elif self.require_first and len(args) > 0 and isinstance(args[0], str):
            obj = read_json(args[0])
            args = (obj, *args[1:])

        return self.call(*args, **kwargs)


class support_dataframe(Decorator):
    """decorator for supporting dataframe
    """
    require_target = True
    target = 'target'

    def wrapper(self, frame, *args, **kwargs):
        if not isinstance(frame, pd.DataFrame):
            return self.call(frame, *args, **kwargs)

        frame = frame.copy()
        if self.require_target and isinstance(args[0], str):
            target = frame.pop(args[0])
            args = (target,) + args[1:]
        elif self.target in kwargs and isinstance(kwargs[self.target], str):
            kwargs[self.target] = frame.pop(kwargs[self.target])

        res = dict()
        for col in frame:
            r = self.call(frame[col], *args, **kwargs)

            if not isinstance(r, np.ndarray):
                r = [r]

            res[col] = r
        return pd.DataFrame(res)

class proxy_docstring(Decorator):
    method_name = None
    
    def __get__(self, *args):
        func = super().__get__(*args)
        
        if self.method_name is not None and hasattr(self._cls, self.method_name):
            setattr(func, '__doc__', getattr(self._cls, self.method_name).__doc__)
        
        return func


import pytest
import numpy as np
import pandas as pd

from .decorator import frame_exclude

np.random.seed(1)


def func():
    "This is a doc for method"


def test_decorator_doc():
    f = frame_exclude(func)

    assert f.__doc__ == 'This is a doc for method'


import sys
import torch
import pytest
import numpy as np
from torch.utils.data import TensorDataset, DataLoader

from .autoencoder import BaseAutoEncoder, VAE

# skip testing with python 3.9 on linux
if sys.version_info >= (3, 9) and sys.platform.startswith('linux'):
    pytest.skip("failed with python 3.9 on linux, need fix!", allow_module_level = True)


X = torch.Tensor(np.random.rand(20000, 784))

loader = DataLoader(
    X,
    batch_size = 128,
    shuffle = True,
)

def test_ae():
    ae = BaseAutoEncoder(784, 200, 10)
    ae.fit(loader, epoch = 1)

def test_vae():
    vae = VAE(784, 200, 10)
    vae.fit(loader, epoch = 1)



from .merge import merge, DTMerge, ChiMerge, StepMerge, QuantileMerge, KMeansMerge
from .detector import detect
from .metrics import KS, KS_bucket, F1
from .stats import quality, IV, VIF, WOE, entropy, entropy_cond, gini, gini_cond
from .selection import select
from .scorecard import ScoreCard
from .version import __version__

VERSION = __version__



import sys
import argparse
import pandas as pd

def func(args):
    """detect csv data

    Examples:

        toad detect -i xxx.csv -o report.csv
    """
    from toad.detector import detect

    sys.stdout.write('reading data....\n')
    with args.input as input:
        data = pd.read_csv(input)

    sys.stdout.write('detecting...\n')
    report = detect(data)

    if args.output:
        sys.stdout.write('saving report...\n')
        report.to_csv(args.output)
        sys.stdout.write('report saved!\n')
    else:
        sys.stdout.write(str(report))
        sys.stdout.write('\n')

    return report

ARGS = {
    'info': {
        'name': 'detect',
        'description': 'detect data from a csv file',
    },
    'defaults': {
        'func': func,
    },
    'args': [
        {
            'flag': ('-i', '--input'),
            'type': argparse.FileType(),
            'help': 'the csv file which will be detected',
            'required': True,
        },
        {
            'flag': ('-o', '--output'),
            'type': argparse.FileType('w'),
            'help': 'path of the csv report will be saved',
        },
    ]
}



import numpy as np
import pandas as pd
from scipy.stats import ks_2samp

from sklearn.metrics import f1_score, roc_auc_score, roc_curve

from .merge import merge

from .utils import (
    feature_splits,
    iter_df,
    unpack_tuple,
    bin_by_splits,
)


def KS(score, target):
    """calculate ks value

    Args:
        score (array-like): list of score or probability that the model predict
        target (array-like): list of real target

    Returns:
        float: the max KS value
    """
    mask = target == 1
    res = ks_2samp(score[mask], score[~mask])
    return res[0]


def KS_bucket(score, target, bucket = 10, method = 'quantile', return_splits = False, **kwargs):
    """calculate ks value by bucket

    Args:
        score (array-like): list of score or probability that the model predict
        target (array-like): list of real target
        bucket (int): n groups that will bin into
        method (str): method to bin score. `quantile` (default), `step`
        return_splits (bool): if need to return splits of bucket

    Returns:
        DataFrame
    """
    df = pd.DataFrame({
        'score': score,
        'bad': target,
    })

    df['good'] = 1 - df['bad']

    bad_total = df['bad'].sum()
    good_total = df['good'].sum()
    all_total = bad_total + good_total

    splits = None
    df['bucket'] = 0

    if bucket is False:
        df['bucket'] = score
    elif isinstance(bucket, (list, np.ndarray, pd.Series)):
        # list of split pointers
        if len(bucket) < len(score):
            bucket = bin_by_splits(score, bucket)
        
        df['bucket'] = bucket
    elif isinstance(bucket, int):
        df['bucket'], splits = merge(score, n_bins = bucket, method = method, return_splits = True, **kwargs)

    grouped = df.groupby('bucket', as_index = False)

    agg1 = pd.DataFrame()
    agg1['min'] = grouped.min()['score']
    agg1['max'] = grouped.max()['score']
    agg1['bads'] = grouped.sum()['bad']
    agg1['goods'] = grouped.sum()['good']
    agg1['total'] = agg1['bads'] + agg1['goods']

    agg2 = (agg1.sort_values(by = 'min')).reset_index(drop = True)

    agg2['bad_rate'] = agg2['bads'] / agg2['total']
    agg2['good_rate'] = agg2['goods'] / agg2['total']

    agg2['odds'] = agg2['bads'] / agg2['goods']

    agg2['bad_prop'] = agg2['bads'] / bad_total
    agg2['good_prop'] = agg2['goods'] / good_total
    agg2['total_prop'] = agg2['total'] / all_total
    

    cum_bads = agg2['bads'].cumsum()
    cum_goods = agg2['goods'].cumsum()
    cum_total = agg2['total'].cumsum()

    cum_bads_rev = agg2.loc[::-1, 'bads'].cumsum()[::-1]
    cum_goods_rev = agg2.loc[::-1, 'goods'].cumsum()[::-1]
    cum_total_rev = agg2.loc[::-1, 'total'].cumsum()[::-1]

    agg2['cum_bad_rate'] = cum_bads / cum_total
    agg2['cum_bad_rate_rev'] = cum_bads_rev / cum_total_rev
    
    agg2['cum_bads_prop'] = cum_bads / bad_total
    agg2['cum_bads_prop_rev'] = cum_bads_rev / bad_total
    agg2['cum_goods_prop'] = cum_goods / good_total
    agg2['cum_goods_prop_rev'] = cum_goods_rev / good_total
    agg2['cum_total_prop'] = cum_total / all_total
    agg2['cum_total_prop_rev'] = cum_total_rev / all_total


    agg2['ks'] = agg2['cum_bads_prop'] - agg2['cum_goods_prop']

    reverse_suffix = ''
    # fix negative ks value
    if agg2['ks'].sum() < 0:
        agg2['ks'] = -agg2['ks']
        reverse_suffix = '_rev'
    
    agg2['lift'] = agg2['bad_prop'] / agg2['total_prop']
    agg2['cum_lift'] = agg2['cum_bads_prop' + reverse_suffix] / agg2['cum_total_prop' + reverse_suffix]

    if return_splits and splits is not None:
        return agg2, splits
    
    return agg2

def KS_by_col(df, by='feature', score='score', target='target'):
    """
    """

    pass


def SSE(y_pred, y):
    """sum of squares due to error
    """
    return np.sum((y_pred - y) ** 2)


def MSE(y_pred, y):
    """mean of squares due to error
    """
    return np.mean((y_pred - y) ** 2)


def AIC(y_pred, y, k, llf = None):
    """Akaike Information Criterion

    Args:
        y_pred (array-like)
        y (array-like)
        k (int): number of featuers
        llf (float): result of log-likelihood function
    """
    if llf is None:
        llf = np.log(SSE(y_pred, y))

    return 2 * k - 2 * llf


def BIC(y_pred, y, k, llf = None):
    """Bayesian Information Criterion

    Args:
        y_pred (array-like)
        y (array-like)
        k (int): number of featuers
        llf (float): result of log-likelihood function
    """
    n = len(y)
    if llf is None:
        llf = np.log(SSE(y_pred, y))

    return np.log(n) * k - 2 * llf


def F1(score, target, split = 'best', return_split = False):
    """calculate f1 value

    Args:
        score (array-like)
        target (array-like)

    Returns:
        float: best f1 score
        float: best spliter
    """
    dataframe = pd.DataFrame({
        'score': score,
        'target': target,
    })

    if split == 'best':
        # find best split for score
        splits = feature_splits(dataframe['score'], dataframe['target'])
    else:
        splits = [split]

    best = 0
    sp = None
    for df, pointer in iter_df(dataframe, 'score', 'target', splits):
        v = f1_score(df['target'], df['score'])

        if v > best:
            best = v
            sp = pointer

    if return_split:
        return best, sp

    return best


def AUC(score, target, return_curve = False):
    """AUC Score

    Args:
        score (array-like): list of score or probability that the model predict
        target (array-like): list of real target
        return_curve (bool): if need return curve data for ROC plot

    Returns:
        float: auc score
    """
    # fix score order
    if np.nanmax(score) > 1:
        score = -score

    auc = roc_auc_score(target, score)

    if not return_curve:
        return auc
    
    return (auc,) + roc_curve(target, score)


def _PSI(test, base):
    test_prop = pd.Series(test).value_counts(normalize = True, dropna = False)
    base_prop = pd.Series(base).value_counts(normalize = True, dropna = False)

    psi = np.sum((test_prop - base_prop) * np.log(test_prop / base_prop))

    frame = pd.DataFrame({
        'test': test_prop,
        'base': base_prop,
    })
    frame.index.name = 'value'

    return psi, frame.reset_index()



def PSI(test, base, combiner = None, return_frame = False):
    """calculate PSI

    Args:
        test (array-like): data to test PSI
        base (array-like): base data for calculate PSI
        combiner (Combiner|list|dict): combiner to combine data
        return_frame (bool): if need to return frame of proportion

    Returns:
        float|Series
    """

    if combiner is not None:
        if isinstance(combiner, (dict, list)):
            from .transform import Combiner
            combiner = Combiner().load(combiner)

        test = combiner.transform(test, labels = True)
        base = combiner.transform(base, labels = True)

    psi = list()
    frame = list()

    if isinstance(test, pd.DataFrame):
        for col in test:
            p, f = _PSI(test[col], base[col])
            psi.append(p)
            frame.append(f)

        psi = pd.Series(psi, index = test.columns)

        frame = pd.concat(
            frame,
            keys = test.columns,
            names = ['columns', 'id'],
        ).reset_index()
        frame = frame.drop(columns = 'id')
    else:
        psi, frame = _PSI(test, base)


    res = (psi,)

    if return_frame:
        res += (frame,)

    return unpack_tuple(res)


def matrix(y_pred, y, splits = None):
    """confusion matrix of target

    Args:
        y_pred (array-like)
        y (array-like)
        splits (float|list): split points of y_pred

    Returns:
        DataFrame: confusion matrix witch true labels in rows and predicted labels in columns

    """
    if splits is not None:
        y_pred = bin_by_splits(y_pred, splits)
    
    labels = np.unique(y)
    from sklearn.metrics import confusion_matrix
    m = confusion_matrix(y, y_pred, labels = labels)

    return pd.DataFrame(
        m,
        index = pd.Index(labels, name = 'Actual'),
        columns = pd.Index(labels, name = 'Predicted'),
    )



import math
import copy
import numpy as np
import pandas as pd
from functools import wraps
from sklearn.base import TransformerMixin
from sklearn.preprocessing import OneHotEncoder
from sklearn.ensemble import GradientBoostingClassifier


from .stats import WOE, probability
from .merge import merge
from .utils.func import to_ndarray, np_count, bin_by_splits, split_target
from .utils.decorator import frame_exclude, select_dtypes
from .utils.mixin import RulesMixin, BinsMixin


class Transformer(TransformerMixin, RulesMixin):
    """Base class for transformers
    """

    _fit_frame = False

    @property
    def _fitted(self):
        return len(self.rules) > 0


    @frame_exclude(is_class = True)
    @select_dtypes(is_class = True)
    def fit(self, X, *args, update = False, **kwargs):
        """fit method, see details in `fit_` method
        """
        dim = getattr(X, 'ndim', 1)

        rules = {}

        if self._fit_frame:
            rules = self.fit_(X, *args, **kwargs)

        elif dim == 1:
            name = getattr(X, 'name', self._default_name)
            rules[name] = self.fit_(X, *args, **kwargs)

        else:
            if len(args) > 0:
                X, y = split_target(X, args[0])
                args = (y, *args[1:])
            if 'y' in kwargs:
                X, kwargs['y'] = split_target(X, kwargs['y'])

            self._check_duplicated_keys(X)

            for col in X:
                name = X[col].name
                rules[name] = self.fit_(X[col], *args, **kwargs)

        if update:
            self.rules.update(rules)
        else:
            self.rules = rules

        return self


    def transform(self, X, *args, **kwargs):
        """transform method, see details in `transform_` method
        """
        if not self._fitted:
            return self._raise_unfitted()


        if self._fit_frame:
            return self.transform_(self.rules, X, *args, **kwargs)

        if getattr(X, 'ndim', 1) == 1:
            if len(self.rules) == 1:
                return self.transform_(self.default_rule(), X, *args, **kwargs)
            elif hasattr(X, 'name') and X.name in self:
                return self.transform_(self.rules[X.name], X, *args, **kwargs)
            else:
                return X

        self._check_duplicated_keys(X)

        res = X.copy()
        for key in X:
            if key in self.rules:
                try:
                    res[key] = self.transform_(self.rules[key], X[key], *args, **kwargs)
                except Exception as e:
                    e.args += ('on column "{key}"'.format(key = key),)
                    raise e

        return res


    def _raise_unfitted(self):
        raise Exception('transformer is unfitted yet!')
    

    def _check_duplicated_keys(self, X):
        if isinstance(X, pd.DataFrame) and X.columns.has_duplicates:
            keys = X.columns[X.columns.duplicated()].values
            raise Exception("X has duplicate keys `{keys}`".format(keys = str(keys)))
        
        return True



class WOETransformer(Transformer):
    """WOE transformer
    """

    def fit_(self, X, y):
        """fit WOE transformer

        Args:
            X (DataFrame|array-like)
            y (str|array-like)
            select_dtypes (str|numpy.dtypes): `'object'`, `'number'` etc. only selected dtypes will be transform
        """
        X = to_ndarray(X)

        value = np.unique(X)
        l = len(value)
        woe = np.zeros(l)

        for i in range(l):
            y_prob, n_prob = probability(y, mask = (X == value[i]))

            woe[i] = WOE(y_prob, n_prob)

        return {
            'value': value,
            'woe': woe,
        }

    def transform_(self, rule, X, default = 'min'):
        """transform function for single feature

        Args:
            X (array-like)
            default (str): 'min'(default), 'max' - the strategy to be used for unknown group

        Returns:
            array-like
        """
        X = to_ndarray(X)
        res = np.zeros(len(X))

        value = rule['value']
        woe = rule['woe']

        if default == 'min':
            default = np.min(woe)
        elif default == 'max':
            default = np.max(woe)

        # replace unknown group to default value
        res[np.isin(X, value, invert = True)] = default

        for i in range(len(value)):
            res[X == value[i]] = woe[i]

        return res

    def _format_rule(self, rule):
        return dict(zip(rule['value'], rule['woe']))

    def _parse_rule(self, rule):
        return {
            'value': np.array(list(rule.keys())),
            'woe': np.array(list(rule.values())),
        }



class Combiner(Transformer, BinsMixin):
    """Combiner for merge data
    """

    def fit_(self, X, y = None, method = 'chi', empty_separate = False, **kwargs):
        """fit combiner

        Args:
            X (DataFrame|array-like): features to be combined
            y (str|array-like): target data or name of target in `X`
            method (str): the strategy to be used to merge `X`, same as `.merge`, default is `chi`
            n_bins (int): counts of bins will be combined
            empty_separate (bool): if need to combine empty values into a separate group
        """
        X = to_ndarray(X)

        if y is not None:
            y = to_ndarray(y)


        if not np.issubdtype(X.dtype, np.number):
            if y is None:
                raise ValueError("Can not combine `{dtype}` type in X, if you want to combine this type columns, please pass argument `y` to deal with it".format(dtype = X.dtype))

            # transform raw data by woe
            transer = WOETransformer()
            woe = transer.fit_transform(X, y)
            # find unique value and its woe value
            uni_val, ix_val = np.unique(X, return_index = True)
            uni_woe = woe[ix_val]
            # sort value by woe
            ix = np.argsort(uni_woe)
            # sort unique value
            uni_val = uni_val[ix]
            # replace X by sorted index
            X = self._raw_to_bin(X, uni_val)

            _, splits = merge(X, target = y, method = method, return_splits = True, **kwargs)

            return self._covert_splits(uni_val, splits)
        

        mask = pd.isna(X)
        if mask.any() and empty_separate:
            X = X[~mask]
            y = y[~mask]
        
        _, splits = merge(X, target = y, method = method, return_splits = True, **kwargs)

        if mask.any() and empty_separate:
            splits = np.append(splits, np.nan)
        
        return splits


    def transform_(self, rule, X, labels = False, ellipsis = 16, **kwargs):
        """transform X by combiner

        Args:
            X (DataFrame|array-like): features to be transformed
            labels (bool): if need to use labels for resulting bins, `False` by default
            ellipsis (int): max length threshold that labels will not be ellipsis, `None` for skipping ellipsis

        Returns:
            array-like
        """
        X = to_ndarray(X)

        # if is not continuous
        if rule.ndim > 1 or not np.issubdtype(rule.dtype, np.number):
            bins = self._raw_to_bin(X, rule)

        else:
            bins = np.zeros(len(X), dtype = int)

            if len(rule):
                # empty to a separate group
                if np.isnan(rule[-1]):
                    mask = pd.isna(X)
                    bins[~mask] = bin_by_splits(X[~mask], rule[:-1])
                    bins[mask] = len(rule)
                else:
                    bins = bin_by_splits(X, rule)

        if labels:
            formated = self.format_bins(rule, index = True, ellipsis = ellipsis)
            empty_mask = (bins == self.EMPTY_BIN)
            bins = formated[bins]
            bins[empty_mask] = self.EMPTY_BIN

        return bins


    def _covert_splits(self, value, splits):
        """covert combine rules to array
        """
        if value is False:
            return splits

        if isinstance(value, np.ndarray):
            value = value.tolist()

        start = 0
        l = list()
        for i in splits:
            i = math.ceil(i)
            l.append(value[start:i])
            start = i

        l.append(value[start:])

        return np.array(l, dtype = object)


    def _raw_to_bin(self, X, splits):
        """bin by splits

        Args:
            X (array-like): feature to be combined
            splits (array-like): splits of `X`

        Returns:
            array-like
        """
        # set default group to EMPTY_BIN
        bins = np.full(X.shape, self.EMPTY_BIN)
        for i in range(len(splits)):
            group = splits[i]
            # if group is else, set all empty group to it
            if isinstance(group, str) and group == self.ELSE_GROUP:
                bins[bins == self.EMPTY_BIN] = i
            else:
                bins[np.isin(X, group)] = i

        return bins


    def set_rules(self, map, reset = False):
        """set rules for combiner

        Args:
            map (dict|array-like): map of splits
            reset (bool): if need to reset combiner

        Returns:
            self
        """
        import warnings
        warnings.warn(
            """`combiner.set_rules` will be deprecated soon,
                use `combiner.load(rules, update = False)` instead!
            """,
            DeprecationWarning,
        )


        self.load(map, update = not reset)

        return self

    def _parse_rule(self, rule):
        return np.array(rule)

    def _format_rule(self, rule, format = False):
        if format:
            rule = self.format_bins(rule)

        return rule.tolist()




class GBDTTransformer(Transformer):
    """GBDT transformer
    """
    _fit_frame = True

    def __init__(self):
        self.gbdt = None
        self.onehot = None


    def fit_(self, X, y, **kwargs):
        """fit GBDT transformer

        Args:
            X (DataFrame|array-like)
            y (str|array-like)
            select_dtypes (str|numpy.dtypes): `'object'`, `'number'` etc. only selected dtypes will be transform,
        """

        if isinstance(y, str):
            X = X.copy()
            y = X.pop(y)

        gbdt = GradientBoostingClassifier(**kwargs)
        gbdt.fit(X, y)

        X = gbdt.apply(X)
        X = X.reshape(-1, X.shape[1])

        onehot = OneHotEncoder().fit(X)

        return {
            'gbdt': gbdt,
            'onehot': onehot,
        }


    def transform_(self, rules, X):
        """transform woe

        Args:
            X (DataFrame|array-like)

        Returns:
            array-like
        """
        X = rules['gbdt'].apply(X)
        X = X.reshape(-1, X.shape[1])
        res = rules['onehot'].transform(X).toarray()
        return res



import sys
import argparse
import pandas as pd

def func(args):
    """detect csv data

    Examples:

        toad evaluate -i xxx.csv
    """
    from .evaluate import evaluate


    sys.stdout.write('reading data....\n')

    test_data = pd.read_csv(args.input)
    if args.base is not None:
        self_data = pd.read_csv(args.base)
    else:
        self_data = None

    arguments = {
        'excel_name': args.name,
        'num': args.top,
        'iv_threshold_value': args.iv,
        'unique_num': args.unique,
        'self_data': self_data,
        'overdue_days': args.overdue,
    }

    evaluate(test_data, **arguments)


ARGS = {
    'info': {
        'name': 'evaluate',
        'description': '第三方数据评估',
    },
    'defaults': {
        'func': func,
    },
    'args': [
        {
            'flag': ('-i', '--input'),
            'type': argparse.FileType('r', encoding='utf-8'),
            'help': '需要评估的 csv 文件',
            'required': True,
        },
        {
            'flag': ('--base',),
            'type': argparse.FileType('r', encoding='utf-8'),
            'help': '用于测试提升效果的基准 csv 数据文件',
            'default': None,
        },
        {
            'flag': ('--overdue',),
            'help': '是否启用逾期天数分析',
            'action': 'store_true',
        },
        {
            'flag': ('--top',),
            'type': int,
            'help': '选择 IV 最高的 n 个变量分析',
            'default': 10,
        },
        {
            'flag': ('--iv',),
            'type': float,
            'help': '选择 IV 大于阈值的变量进行分析',
            'default': 0.02,
        },
        {
            'flag': ('--unique',),
            'type': int,
            'help': '将连续变量合并成 n 组进行分析',
            'default': 10,
        },
        {
            'flag': ('--name',),
            'type': str,
            'help': '生成报告的文件名',
            'default': 'report.xlsx',
        },
    ]
}



import pytest
import numpy as np
import pandas as pd
from sklearn.linear_model import LogisticRegression

from .scorecard import ScoreCard, WOETransformer, Combiner


np.random.seed(1)

ab = np.array(list('ABCDEFG'))
feature = np.random.randint(10, size = 500)
target = np.random.randint(2, size = 500)
str_feat = ab[np.random.choice(7, 500)]

df = pd.DataFrame({
    'A': feature,
    'B': str_feat,
    'C': ab[np.random.choice(2, 500)],
    'D': np.ones(500),
})

card_config = {
    'A': {
        '[-inf ~ 3)': 100,
        '[3 ~ 5)': 200,
        '[5 ~ 8)': 300,
        '[8 ~ inf)': 400,
        'nan': 500,
    },
    'B': {
        ','.join(list('ABCD')): 200,
        ','.join(list('EF')): 400,
        'else': 500,
    },
    'C': {
        'A': 200,
        'B': 100,
    },
}

combiner = Combiner()
bins = combiner.fit_transform(df, target, n_bins = 5)
woe_transer = WOETransformer()
woe = woe_transer.fit_transform(bins, target)



# create a score card
card = ScoreCard(
    combiner = combiner,
    transer = woe_transer,
)
card.fit(woe, target)

FUZZ_THRESHOLD = 1e-4
TEST_SCORE = pytest.approx(453.58, FUZZ_THRESHOLD)


def test_representation():
    repr(card)

def test_load():
    card = ScoreCard().load(card_config)
    score = card.predict(df)
    assert score[200] == 600

def test_load_after_init_combiner():
    card = ScoreCard(
        combiner = combiner,
        transer = woe_transer,
    )
    card.load(card_config)
    score = card.predict(df)
    assert score[200] == 600

def test_proba_to_score():
    model = LogisticRegression()
    model.fit(woe, target)

    proba = model.predict_proba(woe)[:,1]
    score = card.proba_to_score(proba)
    assert score[404] == TEST_SCORE

def test_score_to_prob():
    score = card.predict(df)
    proba = card.score_to_proba(score)
    assert proba[404] == 0.4673929989138551

def test_predict():
    score = card.predict(df)
    assert score[404] == TEST_SCORE

def test_predict_proba():
    proba = card.predict_proba(df)
    assert proba[404,1] == 0.4673929989138551

def test_predict_sub_score():
    score, sub = card.predict(df, return_sub = True)
    assert sub.loc[250, 'B'] == pytest.approx(162.0781460573475, FUZZ_THRESHOLD)

def test_woe_to_score():
    score = card.woe_to_score(woe)
    score = np.sum(score, axis = 1)
    assert score[404] == TEST_SCORE

def test_bin_to_score():
    score = card.bin_to_score(bins)
    assert score[404] == TEST_SCORE

def test_export_map():
    card_map = card.export()
    assert card_map['B']['D'] == 159.24

def test_card_map():
    config = card.export()
    card_from_map = ScoreCard().load(config)
    score = card_from_map.predict(df)
    assert score[404] == TEST_SCORE

def test_card_map_with_else():
    card_from_map = ScoreCard().load(card_config)
    score = card_from_map.predict(df)
    assert score[80] == 1000

def test_generate_testing_frame():
    card = ScoreCard().load(card_config)
    frame = card.testing_frame()
    assert frame.loc[4, 'B'] == 'E'

def test_export_frame():
    card = ScoreCard().load(card_config)
    frame = card.export(to_frame = True)
    rows = frame[(frame['name'] == 'B') & (frame['value'] == 'else')].reset_index()
    assert rows.loc[0, 'score'] == 500

def test_card_combiner_number_not_match():
    c = combiner.export()
    c['A'] = [0, 3, 6, 8]
    com = Combiner().load(c)
    bins = com.transform(df)
    woe_transer = WOETransformer()
    woe = woe_transer.fit_transform(bins, target)

    card = ScoreCard(
        combiner = com,
        transer = woe_transer,
    )

    with pytest.raises(Exception) as e:
        # will raise an exception when fitting a card
        card.fit(woe, target)

    assert '\'A\' is not matched' in str(e.value)


def test_card_combiner_str_not_match():
    c = combiner.export()
    c['C'] = [['A'], ['B'], ['C']]
    com = Combiner().load(c)
    bins = com.transform(df)
    woe_transer = WOETransformer()
    woe = woe_transer.fit_transform(bins, target)

    card = ScoreCard(
        combiner = com,
        transer = woe_transer,
    )

    with pytest.raises(Exception) as e:
        # will raise an exception when fitting a card
        card.fit(woe, target)

    assert '\'C\' is not matched' in str(e.value)


def test_card_with_less_X():
    x = woe.drop(columns = 'A')
    card = ScoreCard(
        combiner = combiner,
        transer = woe_transer,
    )

    card.fit(x, target)
    assert card.predict(x)[200] == pytest.approx(411.968588097131, FUZZ_THRESHOLD)




import re
import json
import string
import numpy as np
import pandas as pd
from functools import wraps

from multiprocessing import Pool, current_process, cpu_count


CONTINUOUS_NUM = 10
FEATURE_THRESHOLD = 1e-7
NAN_REPLACEMENT = -2e10


NAN_LIST = [
    'nan',
    'Nan',
    'null',
    'None',
    None,
    np.nan,
]


class Parallel:
    def __init__(self):
        self.ismain = False
        self.results = []
        self.pro = current_process()

        if self.pro.name == 'MainProcess':
            self.ismain = True
            self.pool = Pool(cpu_count())


    def apply(self, func, args = (), kwargs = {}):
        if not self.ismain:
            r = func(*args, **kwargs)
        else:
            r = self.pool.apply_async(func, args = args, kwds = kwargs)

        self.results.append(r)

    def join(self):
        if not self.ismain:
            return self.results

        self.pool.close()
        self.pool.join()

        return [r.get() for r in self.results]



def np_count(arr, value, default = None):
    c = (arr == value).sum()

    if default is not None and c == 0:
        return default

    return c


def _replace_nan(arr):
    a = np.copy(arr)
    a[a == NAN_REPLACEMENT] = np.nan
    return a


def has_nan(arr):
    return np.any(pd.isna(arr))


def np_unique(arr, **kwargs):
    arr = to_ndarray(arr)

    if not has_nan(arr):
        return np.unique(arr, **kwargs)

    arr[np.isnan(arr)] = NAN_REPLACEMENT

    res = np.unique(arr, **kwargs)

    if isinstance(res, tuple):
        u = _replace_nan(res[0])
        return (u, *res[1:])

    return _replace_nan(res)


def to_ndarray(s, dtype = None):
    """
    """
    if isinstance(s, np.ndarray):
        arr = np.copy(s)
    elif isinstance(s, pd.core.base.PandasObject):
        arr = np.copy(s.values)
    else:
        arr = np.array(s)


    if dtype is not None:
        arr = arr.astype(dtype)
    # covert object type to str
    elif arr.dtype.type is np.object_:
        arr = arr.astype(np.str)

    return arr


def fillna(feature, by = -1):
    # copy array
    copied = np.copy(feature)

    mask = pd.isna(copied)

    copied[mask] = by

    return copied

def bin_by_splits(feature, splits):
    """Bin feature by split points
    """
    feature = fillna(feature)

    if not isinstance(splits, (list, np.ndarray)):
        splits = [splits]
    
    return np.digitize(feature, splits)


def feature_splits(feature, target):
    """find posibility spilt points
    """
    feature = to_ndarray(feature)
    target = to_ndarray(target)

    matrix = np.vstack([feature, target])
    matrix = matrix[:, matrix[0,:].argsort()]

    splits_values = []
    for i in range(1, len(matrix[0])):
        # if feature value is almost same, then skip
        if matrix[0,i] <= matrix[0, i-1] + FEATURE_THRESHOLD:
            continue

        # if target value is not same, calculate split
        if matrix[1, i] != matrix[1, i-1]:
            v = (matrix[0, i] + matrix[0, i-1]) / 2.0
            splits_values.append(v)

    return np.unique(splits_values)


def iter_df(dataframe, feature, target, splits):
    """iterate dataframe by split points

    Returns:
        iterator (df, splitter)
    """
    splits.sort()
    df = pd.DataFrame()
    df['source'] = dataframe[feature]
    df[target] = dataframe[target]
    df[feature] = 0

    for v in splits:
        df.loc[df['source'] < v, feature] = 1
        yield df, v


def inter_feature(feature, splits):
    splits.sort()
    bin = np.zeros(len(feature))

    for v in splits:
        bin[feature < v] = 1
        yield bin


def is_continuous(series):
    series = to_ndarray(series)
    if not np.issubdtype(series.dtype, np.number):
        return False

    n = len(np.unique(series))
    return n > CONTINUOUS_NUM or n / series.size > 0.5
    # return n / series.size > 0.5


def split_target(frame, target):
    """
    """
    if isinstance(target, str):
        cols = frame.columns.copy().drop(target)
        f = frame[cols]
        t = frame[target]
    else:
        f = frame.copy(deep = False)
        t = target

    return f, t


def unpack_tuple(x):
    if len(x) == 1:
        return x[0]
    else:
        return x

ALPHABET = string.ascii_uppercase + string.digits
def generate_str(size = 6, chars = ALPHABET):
    return ''.join(np.random.choice(list(chars), size = size))



def save_json(contents, file, indent = 4):
    """save json file

    Args:
        contents (dict): contents to save
        file (str|IOBase): file to save
    """
    if isinstance(file, str):
        file = open(file, 'w')

    with file as f:
        json.dump(contents, f, ensure_ascii = False, indent = indent)


def read_json(file):
    """read json file
    """
    if isinstance(file, str):
        file = open(file)

    with file as f:
        res = json.load(f)

    return res




def clip(series, value = None, std = None, quantile = None):
    """clip series

    Args:
        series (array-like): series need to be clipped
        value (number | tuple): min/max value of clipping
        std (number | tuple): min/max std of clipping
        quantile (number | tuple): min/max quantile of clipping
    """
    series = to_ndarray(series)

    if value is not None:
        min, max = _get_clip_value(value)

    elif std is not None:
        min, max = _get_clip_value(std)
        s = np.std(series, ddof = 1)
        mean = np.mean(series)
        min = None if min is None else mean - s * min
        max = None if max is None else mean + s * max

    elif quantile is not None:
        if isinstance(quantile, tuple):
            min, max = quantile
        else:
            min = quantile
            max = 1 - quantile

        min = None if min is None else np.quantile(series, min)
        max = None if max is None else np.quantile(series, max)

    else:
        return series

    return np.clip(series, min, max)


def _get_clip_value(params):
    if isinstance(params, tuple):
        return params
    else:
        return params, params


def diff_time(base, target, format = None, time = 'day'):
    # if base is not a datetime list
    if not np.issubdtype(base.dtype, np.datetime64):
        base = pd.to_datetime(base, format = format, cache = True)

    target = pd.to_datetime(target, format = format, cache = True)

    delta = target - base

    if time == 'day':
        return delta.dt.days

    return delta


def diff_time_frame(base, frame, format = None):
    res = pd.DataFrame()

    base = pd.to_datetime(base, format = format, cache = True)

    for col in frame:
        try:
            res[col] = diff_time(base, frame[col], format = format)
        except Exception as e:
            continue

    return res


def flatten_columns(columns, sep = '_'):
    """flatten multiple columns to 1-dim columns joined with '_'
    """
    l = []
    for col in columns:
        if not isinstance(col, str):
            col = sep.join(col)
        
        l.append(col)
    
    return l


def bin_to_number(reg = None):
    """
    Returns:
        function: func(string) -> number
    """
    if reg is None:
        reg = r'\d+'

    def func(x):
        if pd.isnull(x):
            return np.nan

        res = re.findall(reg, x)
        l = len(res)
        res = map(float, res)
        if l == 0:
            return np.nan
        else:
            return sum(res) / l

    return func


def generate_target(size, rate = 0.5, weight = None, reverse = False):
    """generate target for reject inference

    Args:
        size (int): size of target
        rate (float): rate of '1' in target
        weight (array-like): weight of '1' to generate target
        reverse (bool): if need reverse weight

    Returns:
        array
    """
    if weight is not None:
        weight = np.asarray(weight)

        if reverse is True:
            weight = (np.max(weight) + np.min(weight)) - weight

        weight = weight / weight.sum()

    res = np.zeros(size)

    choice_num = int(size * rate)
    ix = np.random.choice(size, choice_num, replace = False, p = weight)
    res[ix] = 1

    return res


def get_dummies(dataframe, exclude = None, binary_drop = False, **kwargs):
    """get dummies
    """
    columns = dataframe.select_dtypes(exclude = 'number').columns

    if len(columns) == 0:
        return dataframe

    if exclude is not None:
        columns = columns.difference(exclude)

    if binary_drop:
        mask = dataframe[columns].nunique(dropna = False) == 2

        if mask.sum() != 0:
            dataframe = pd.get_dummies(
                dataframe,
                columns = columns[mask],
                drop_first = True,
                **kwargs,
            )
            columns = columns[~mask]

    data = pd.get_dummies(dataframe, columns = columns, **kwargs)
    return data



import os
import pkgutil
from importlib import import_module

COMMAND_DIR = os.path.dirname(os.path.abspath(__file__))

def get_plugins():
    plugins = []

    for _, name, ispkg in pkgutil.iter_modules([COMMAND_DIR]):
        if ispkg:
            module = import_module('toad.commands.{}'.format(name))
            plugins.append(module)

    return plugins



import pytest
import numpy as np
import pandas as pd

import pyximport

pyximport.install(setup_args={"include_dirs": np.get_include()})

from toad.cli import get_parser

def disable_stdout(fn):

    def wrapper(*args):
        import os
        import sys

        with open(os.devnull, 'w') as f:
            so = sys.stdout
            sys.stdout = f

            fn(*args)

            sys.stdout = so

    return wrapper


parser = get_parser()



@disable_stdout
def test_detect():
    args = parser.parse_args(['detect', '-i', 'tests/test_data.csv'])
    rep = args.func(args)
    assert rep.loc['E', 'unique'] == 20

@pytest.mark.skip("tree command will generate a pic in travis-ci log")
@disable_stdout
def test_tree():
    args = parser.parse_args(['tree', '-i', 'tests/test_data.csv'])
    args.func(args)
    pass



import numpy as np
from torch import optim

from .history import History
from .earlystopping import EarlyStopping
from ...utils.progress import Progress
        

        
class Trainer:
    def __init__(self, model, loader, optimizer = None, keep_history = None,
                early_stopping = EarlyStopping()):
        """
        Args:
            model (nn.Module)
            loader (torch.DataLoader)
            optimizer (torch.Optimier)
            early_stopping (EarlyStopping)
            keep_history (int): keep the last n-th epoch logs, `None` will keep all
        """
        self.model = model
        self.loader = loader

        if optimizer is None:
            optimizer = optim.Adam(model.parameters(), lr = 1e-3)
        
        self.optimizer = optimizer

        self.early_stop = early_stopping

        from collections import deque
        self.history = deque(maxlen = keep_history)


    def train(self, epoch = 10, callback = None):
        """
        Args:
            epoch (int): number of epoch for training loop
            callback (callable): callable function will be called every epoch
        """
        # init progress bar
        p = Progress(self.loader)

        

        for ep in range(epoch):
            p.prefix = f"Epoch:{ep}"

            # setup a new history for model in each epoch
            history = History()
            self.history.append(history)
            self.model._history = history

            loss = 0.
            for i, batch in enumerate(p, start = 1):
                # step fit
                l = self.model.fit_step(batch)

                # log loss
                self.model.log('loss', l)
                
                self.optimizer.zero_grad()
                l.backward()
                self.optimizer.step()

                loss += (l.item() - loss) / i
                p.suffix = 'loss:{:.4f}'.format(loss)


            if self.early_stop and self.early_stop(self.model, history, epoch = ep):
                # set best state to model
                best_state = self.early_stop.get_best_state()
                self.model.load_state_dict(best_state)
                break
            
            if callable(callback):
                callback(ep, history)
        
        return self.model



import numpy as np
import pandas as pd

from .stats import IV
from .metrics import AUC
from .tadpole import tadpole
from .tadpole.utils import HEATMAP_CMAP, MAX_STYLE, add_annotate, add_text, reset_ylim
from .utils import unpack_tuple, generate_str

def badrate_plot(frame, x = None, target = 'target', by = None,
                freq = None, format = None, return_counts = False,
                return_proportion = False, return_frame = False):
    """plot for badrate

    Args:
        frame (DataFrame)
        x (str): column in frame that will be used as x axis
        target (str): target column in frame
        by (str): column in frame that will be calculated badrate by it
        freq (str): offset aliases string by pandas
                    http://pandas.pydata.org/pandas-docs/stable/timeseries.html#offset-aliases
        format (str): format string for time
        return_counts (bool): if need return counts plot
        return_frame (bool): if need return frame

    Returns:
        Axes: badrate plot
        Axes: counts plot
        Axes: proportion plot
        Dataframe: grouping detail data
    """
    frame = frame.copy()
    markers = True

    if not isinstance(target, str):
        temp_name = generate_str()
        frame[temp_name] = target
        target = temp_name

    grouper = x
    if freq is not None:
        frame.loc[:, x] = pd.to_datetime(frame[x], format = format)
        grouper = pd.Grouper(key = x, freq = freq)

    if by is not None:
        grouper = [by, grouper]

        styles_count = frame[by].nunique()
        if styles_count > MAX_STYLE:
            markers = ['o'] * styles_count

    group = frame.groupby(grouper)
    table = group[target].agg(['sum', 'count']).reset_index()
    table['badrate'] = table['sum'] / table['count']

    # set number dtype to object
    if np.issubdtype(table[x].dtype, np.number):
        table[x] = table[x].astype(str)


    rate_plot = tadpole.lineplot(
        x = x,
        y = 'badrate',
        hue = by,
        style = by,
        data = table,
        legend = 'full',
        markers = markers,
        dashes = False,
    )

    # set y axis start with 0
    rate_plot.set_ylim(0, None)

    res = (rate_plot,)

    if return_counts:
        count_plot = tadpole.barplot(
            x = x,
            y = 'count',
            hue = by,
            data = table,
        )
        res += (count_plot,)


    if return_proportion:
        table['prop'] = 0
        for v in table[x].unique():
            mask = (table[x] == v)
            table.loc[mask, 'prop'] = table[mask]['count'] / table[mask]['count'].sum()

        prop_plot = tadpole.barplot(
            x = x,
            y = 'prop',
            hue = by,
            data = table,
        )
        res += (prop_plot,)


    if return_frame:
        res += (table,)

    return unpack_tuple(res)


def corr_plot(frame, figure_size = (20, 15)):
    """plot for correlation

    Args:
        frame (DataFrame): frame to draw plot
    Returns:
        Axes
    """
    corr = frame.corr()

    mask = np.zeros_like(corr, dtype = np.bool)
    mask[np.triu_indices_from(mask)] = True

    map_plot = tadpole.heatmap(
        corr,
        mask = mask,
        cmap = HEATMAP_CMAP,
        vmax = 1,
        vmin = -1,
        center = 0,
        square = True,
        cbar_kws = {"shrink": .5},
        linewidths = .5,
        annot = True,
        fmt = '.2f',
        figure_size = figure_size,
    )

    return map_plot


def proportion_plot(x = None, keys = None):
    """plot for comparing proportion in different dataset

    Args:
        x (Series|list): series or list of series data for plot
        keys (str|list): keys for each data

    Returns:
        Axes
    """
    if not isinstance(x, list):
        x = [x]

    if keys is None:
        keys = [
            x[ix].name
            if hasattr(x[ix], 'name') and x[ix].name is not None
            else ix
            for ix in range(len(x))
        ]
    elif isinstance(keys, str):
        keys = [keys]

    x = map(pd.Series, x)
    data = pd.concat(x, keys = keys, names = ['keys']).reset_index()
    data = data.rename(columns = {data.columns[2]: 'value'})

    prop_data = data.groupby('keys')['value'].value_counts(
        normalize = True,
        dropna = False,
    ).rename('proportion').reset_index()

    prop_plot = tadpole.barplot(
        x = 'value',
        y = 'proportion',
        hue = 'keys',
        data = prop_data,
    )

    return prop_plot


def roc_plot(score, target, compare = None):
    """plot for roc

    Args:
        score (array-like): predicted score
        target (array-like): true target
        compare (array-like): another score for comparing with score

    Returns:
        Axes
    """
    auc, fpr, tpr, thresholds = AUC(score, target, return_curve = True)

    ax = tadpole.lineplot(
        x = fpr,
        y = tpr,
    )
    ax.fill_between(fpr, tpr, alpha = 0.3)
    ax = add_text(ax, 'AUC: {:.5f}'.format(auc))

    if compare is not None:
        c_aux, c_fpr, c_tpr, _ = AUC(compare, target, return_curve = True)
        ax.plot(c_fpr, c_tpr)
        ax.fill_between(c_fpr, c_tpr, alpha = 0.3)

    ax.plot([0, 1], [0, 1], color = 'red', linestyle = '--')

    return ax


def bin_plot(frame, x = None, target = 'target', iv = True, annotate_format = ".2f"):
    """plot for bins

    Args:
        frame (DataFrame)
        x (str): column in frame that will be used as x axis
        target (str): target column in frame
        iv (bool): if need to show iv in plot
        annotate_format (str): format str for axis annotation of chart

    Returns:
        Axes: bins' proportion and badrate plot
    """
    frame = frame.copy()

    if not isinstance(target, str):
        temp_name = generate_str()
        frame[temp_name] = target
        target = temp_name
    
    
    group = frame.groupby(x)

    table = group[target].agg(['sum', 'count']).reset_index()
    table['badrate'] = table['sum'] / table['count']
    table['prop'] = table['count'] / table['count'].sum()

    prop_ax = tadpole.barplot(
        x = x,
        y = 'prop',
        data = table,
        color = '#82C6E2',
    )

    prop_ax = add_annotate(prop_ax, format = annotate_format)

    badrate_ax = prop_ax.twinx()
    badrate_ax.grid(False)

    badrate_ax = tadpole.lineplot(
        x = x,
        y = 'badrate',
        data = table,
        color = '#D65F5F',
        ax = badrate_ax,
    )

    badrate_ax.set_ylim([0, None])
    badrate_ax = add_annotate(badrate_ax, format = annotate_format)

    if iv:
        prop_ax = reset_ylim(prop_ax)
        prop_ax = add_text(prop_ax, 'IV: {:.5f}'.format(IV(frame[x],frame[target])))

    return prop_ax



"""
Windows:
    conda install python-graphviz
Mac:
    brew install graphviz
    pip install graphviz
"""

import numpy as np
import pandas as pd

import graphviz

import sklearn
from sklearn.tree import DecisionTreeClassifier


def tree_to_dot(tree, features, high_light = 0.15):
    from io import StringIO
    from sklearn.tree import _tree

    out = StringIO()
    tree_ = tree.tree_

    features = np.array([
        features[i] if i != _tree.TREE_UNDEFINED else "undefined!"
        for i in tree_.feature
    ])

    out.write('digraph Tree {\n')
    out.write('edge [fontname="FangSong"];\n')
    out.write('node [shape=box];\n')

    def recurse(node, parent = None, label = None):
        sample = tree_.n_node_samples[node]
        bad_rate = tree_.value[node][0,1] / sample

        out.write('{} [label="'.format(node))

        out.write('bad rate: {:.2%}\n'.format(bad_rate))
        out.write('sample: {:.2%}\n'.format(sample / tree_.n_node_samples[0]))

        # end of label
        out.write('"')

        if bad_rate > high_light:
            out.write(', color="red"')

        # end of node
        out.write('];\n')

        if tree_.feature[node] != _tree.TREE_UNDEFINED:
            name = features[node]
            threshold = tree_.threshold[node]
            recurse(tree_.children_left[node], node, '{} <= {:.2f}'.format(name, threshold))
            recurse(tree_.children_right[node], node, '{} > {:.2f}'.format(name, threshold))

        if parent is not None:
            out.write('{} -> {} [label="{}"];\n'.format(parent, node, label))

    recurse(0, None)

    out.write('}')
    s = out.getvalue()
    out.close()
    return s


def dot_to_img(dot, file = 'report.png'):
    import os

    name, ext = os.path.splitext(file)

    graph = graphviz.Source(dot)
    graph.format = ext[1:]
    graph.view(name, cleanup = True)


def split_data(frame, target = 'target'):
    X = frame.drop(columns = target)

    res = (X,)
    if isinstance(target, str):
        target = [target]

    for col in target:
        res += (frame[col],)

    return res


def dtree(frame, target, criterion = 'gini', depth = None, sample = 0.01, ratio = 0.15):
    tree = DecisionTreeClassifier(
        criterion = criterion,
        min_samples_leaf = sample,
        max_depth = depth,
    )

    tree.fit(frame.fillna(-1), target)

    dot_string = tree_to_dot(tree, frame.columns.values, high_light = ratio)

    dot_to_img(dot_string, file = target.name + '.png')



from time import time

class Progress:
    def __init__(self, iterable, size = None):
        self.iterable = iterable

        self.batch = 1
        self.size = size
        if hasattr(iterable, '__len__'):
            self.size = len(iterable)
        
        # is pytorch dataloader
        if hasattr(iterable, 'batch_size'):
            self.batch = getattr(iterable, 'batch_size')
            self.size = len(iterable.dataset)


        self.idx = 0
        self.time = None


        self.BAR_LENGTH = 32
        
        self.SYMBOL_DONE = '█'
        self.SYMBOL_REST = '.'
        self.prefix = ""
        self.suffix = ""
        self.template = "{prefix} {percent:3.0f}%|{bar}| [{done}/{size}] {time:.2f}s {suffix}"


    def __len__(self):
        return self.size
    

    def __iter__(self):
        start = time()
        for item in self.iterable:
            yield item

            self.time = time() - start
            self.idx += 1
            self.flush()
        
        # reset index
        self.idx = 0
        print()

    def flush(self):
        done = min(self.idx * self.batch, self.size)
        percent = done / self.size

        bar = (self.SYMBOL_DONE * int(percent * self.BAR_LENGTH)).ljust(self.BAR_LENGTH, self.SYMBOL_REST)

        print('\r' + self.template.format(
            percent = percent * 100,
            bar = bar,
            done = done,
            size = self.size,
            time = self.time,
            prefix = self.prefix,
            suffix = self.suffix,
        ), end = '')






import pytest
import numpy as np
import pandas as pd

from .metrics import KS, KS_bucket, F1, PSI, AUC, matrix

np.random.seed(1)

feature = np.random.rand(500)
target = np.random.randint(2, size = 500)
base_feature = np.random.rand(500)

test_df = pd.DataFrame({
    'A': np.random.rand(500),
    'B': np.random.rand(500),
})
base_df = pd.DataFrame({
    'A': np.random.rand(500),
    'B': np.random.rand(500),
})

FUZZ_THRESHOLD = 1e-10

def test_KS():
    result = KS(feature, target)
    assert result == 0.05536775661256989

def test_KS_bucket():
    result = KS_bucket(feature, target)
    assert result.loc[4, 'ks'] == -0.028036335090276976

def test_KS_bucket_use_step():
    result = KS_bucket(feature, target, method = 'step', clip_q = 0.01)
    assert result.loc[4, 'ks'] == -0.0422147102645028

def test_KS_bucket_for_all_score():
    result = KS_bucket(feature, target, bucket = False)
    assert len(result) == 500

def test_KS_bucket_return_splits():
    result, splits = KS_bucket(feature, target, return_splits = True)
    assert len(splits) == 9

def test_KS_bucket_use_split_pointers():
    result = KS_bucket(feature, target, bucket = [0.2, 0.6])
    assert len(result) == 3

def test_KS_bucket_with_lift():
    result = KS_bucket(feature, target)
    assert result.loc[3, 'lift'] == 1.0038610038610036

def test_KS_bucket_with_cum_lift():
    result = KS_bucket(feature, target)
    assert result.loc[3, 'cum_lift'] == 1.003861003861004


def test_F1():
    result, split = F1(feature, target, return_split = True)
    assert result == 0.6844207723035951

def test_F1_split():
    result = F1(feature, target, split = 0.5)
    assert result == 0.51417004048583

def test_AUC():
    result = AUC(feature, target)
    assert result == 0.5038690142424582

def test_AUC_with_curve():
    auc, fpr, tpr, thresholds = AUC(feature, target, return_curve = True)
    assert thresholds[200] == 0.15773006987053328

def test_PSI():
    result = PSI(feature, base_feature, combiner = [0.3, 0.5, 0.7])
    assert result == 0.018630024627491467

def test_PSI_frame():
    result = PSI(
        test_df,
        base_df,
        combiner = {
            'A': [0.3, 0.5, 0.7],
            'B': [0.4, 0.8],
        },
    )

    assert result['B'] == pytest.approx(0.014528279995858708, FUZZ_THRESHOLD)

def test_PSI_return_frame():
    result, frame = PSI(
        test_df,
        base_df,
        combiner = {
            'A': [0.3, 0.5, 0.7],
            'B': [0.4, 0.8],
        },
        return_frame = True,
    )

    assert frame.loc[4, 'test'] == 0.38

def test_matrix():
    df = matrix(feature, target, splits = 0.5)
    assert df.iloc[0,1] == 133



import torch
import numpy as np

class History:
    """model history
    """
    def __init__(self):
        self._store = {}
    

    def __getitem__(self, key):
        return self._store[key]
    

    def __setitem__(self, key, value):
        return self.log(key, value)
    

    def _push(self, key, value):
        """push value into history

        Args:
            key (str): key of history
            value (np.ndarray): an array of values
        """
        if key not in self._store:
            self._store[key] = value
            return

        self._store[key] = np.concatenate([
            self._store[key],
            value,
        ])
    

    def log(self, key, value):
        """log message to history

        Args:
            key (str): name of message
            value (Tensor): tensor of values
        """
        if isinstance(value, torch.Tensor):
            value = value.detach().cpu().numpy()
            
            # fix scaler tensor
            if value.ndim == 0:
                value = value.reshape(-1)

        elif np.isscalar(value):
            value = np.array([value])
        else:
            raise TypeError("value should be `torch.Tensor` or `scalar`")
        
        self._push(key, value)



import torch
from torch import nn
from torch.nn.functional import relu, binary_cross_entropy

from ..module import Module



class BaseAutoEncoder(Module):
    def __init__(self, input, hidden, zipped):
        super().__init__()

        self.encoder = nn.Sequential(
            nn.Linear(input, hidden),
            nn.ReLU(),
            nn.Linear(hidden, zipped),
        )

        self.decoder = nn.Sequential(
            nn.Linear(zipped, hidden),
            nn.ReLU(),
            nn.Linear(hidden, input),
        )

        self.loss = nn.MSELoss()
    
    
    def encode(self, x):
        return self.encoder(x)
    
    def decode(self, x):
        return self.decoder(x)

    def forward(self, x):
        z = self.encode(x)
        return self.decode(z)
    
    def fit_step(self, x):
        return self.loss(self(x), x)



class VAE(Module):
    def __init__(self, input, hidden, zipped):
        super().__init__()
        
        self.hidden_layer = nn.Linear(input, hidden)

        self.mu_layer = nn.Linear(hidden, zipped)
        self.var_layer = nn.Linear(hidden, zipped)

        self.decoder = nn.Sequential(
            nn.Linear(zipped, hidden),
            nn.ReLU(),
            nn.Linear(hidden, input),
        )

        self.loss = nn.MSELoss()
    
    def encode(self, x):
        h = relu(self.hidden_layer(x))
        mu = self.mu_layer(h)
        var = self.var_layer(h)

        std = torch.exp(var / 2)
        eps = torch.rand_like(std)
        
        z = mu + eps * std
        return z, mu, var
    
    def decode(self, x):
        return self.decoder(x)
    
    def forward(self, x):
        z, mu, var = self.encode(x)
        x_hat = self.decode(z)
        return x_hat, mu, var
    
    def fit_step(self, x):
        x_hat, mu, var = self(x)
        l = self.loss(x_hat, x)
        kld = -0.5 * torch.sum(1 + var - torch.pow(mu, 2) - torch.exp(var))

        loss = l + kld
        return loss



# cython: language_level = 3, infer_types = True, boundscheck = False

import numpy as np
cimport numpy as np
cimport cython



cdef number c_min(number[:] arr):
    cdef number res = np.inf

    for i in range(arr.shape[0]):
        if res > arr[i]:
            res = arr[i]
    return res


cdef number c_sum(number[:,:] arr):
    cdef number res = 0

    cdef Py_ssize_t i,j
    for i in range(arr.shape[0]):
        for j in range(arr.shape[1]):
            res += arr[i, j]

    return res


cdef number[:] c_sum_axis_0(number[:,:] arr):
    cdef number[:] res = np.zeros(arr.shape[1], dtype=np.float)

    for i in range(arr.shape[0]):
        for j in range(arr.shape[1]):
            res[j] += arr[i, j]

    return res


cdef number[:] c_sum_axis_1(number[:,:] arr):
    cdef number[:] res = np.zeros(arr.shape[0], dtype=np.float)

    for i in range(arr.shape[0]):
        for j in range(arr.shape[1]):
            res[i] += arr[i, j]

    return res



import torch
from torch import nn
import torch.nn.functional as F
from torch.utils.data import TensorDataset, DataLoader

from .module import Module

DATASET_SIZE = 20000
NUM_FEATS = 784
NUM_CLASSES = 2

X = torch.rand(DATASET_SIZE, NUM_FEATS, dtype = torch.float)
y = torch.randint(NUM_CLASSES, size = (DATASET_SIZE,), dtype = torch.long)

loader = DataLoader(
    TensorDataset(X, y),
    batch_size = 128,
    shuffle = True,
)

class TestModel(Module):
    def __init__(self, in_feats, out_feats):
        super().__init__()
        
        self.linear = nn.Linear(in_feats, out_feats)
    
    def forward(self, x):
        x = self.linear(x)
        return F.relu(x)
    
    def fit_step(self, batch):
        x, y = batch
        y_hat = self(x)
        return F.cross_entropy(y_hat, y)

def test_model():
    model = TestModel(NUM_FEATS, NUM_CLASSES)
    model.fit(loader, epoch = 1)


def test_fit_callback():
    history = []

    def func(e, logs):
        history.append(logs)
    
    model = TestModel(NUM_FEATS, NUM_CLASSES)
    model.fit(loader, epoch = 2, callback = func)
    assert len(history) == 2



import pytest
import numpy as np
import pandas as pd


from .process import Processing, Mask, F


np.random.seed(1)

history = np.full(500, np.datetime64('2020-03-01')) - np.random.randint(30, 400, size = 500)
open_time = np.full(500, np.datetime64('2020-03-01')) - np.random.randint(30, size = 500)
A = np.random.randint(10, size = 500)
B = np.random.rand(500)
B[np.random.choice(500, 10)] = np.nan


df = pd.DataFrame({
    'history': history,
    'open_time': open_time,
    'A': A,
    'B': B,
})


def test_mask():
    m = Mask('A') > 3
    assert m.replay(df).sum() == (A > 3).sum()


def test_mask_without_name():
    m = Mask() > 3
    assert m.replay(A).sum() == (A > 3).sum()

def test_mask_isin():
    m = Mask('A').isin([1,2,3])
    assert m.replay(df).sum() == df['A'].isin([1,2,3]).sum()

def test_mask_isna():
    m = Mask('A').isna()
    assert m.replay(df).sum() == df['A'].isna().sum()

def test_f():
    assert F(len)(A)[0] == 500

def test_processing():
    res = (
        Processing(df)
        .groupby('open_time')
        .apply({'A': ['min', 'mean']})
        .apply({'B': [
            {
                'f': 'size',
                'mask': Mask('A') > 1,
            },
            {
                'f': len,
            },
        ]})
        .exec()
    )
    
    assert res.size == 120 and res.loc['2020-02-29', 'B_size'] == 23


def test_processing_with_partition():
    from .partition import ValuePartition
    res = (
        Processing(df)
        .groupby('open_time')
        .partitionby(ValuePartition('A'))
        .apply({'B': ['mean', 'size']})
        .exec()
    )

    assert res.size == 600 and res.loc['2020-02-29', 'B_size_1'] == 2


import pytest
import numpy as np
import pandas as pd

import pyximport

pyximport.install(setup_args={"include_dirs": np.get_include()})

from .transform import WOETransformer, Combiner, GBDTTransformer

np.random.seed(1)

ab = np.array(list('ABCDEFG'))
feature = np.random.randint(10, size = 500)
target = np.random.randint(2, size = 500)
str_feat = ab[np.random.choice(7, 500)]
uni_feat = np.ones(500)
empty_feat = feature.astype(float)
empty_feat[np.random.choice(500, 50, replace = False)] = np.nan

df = pd.DataFrame({
    'A': feature,
    'B': str_feat,
    'C': uni_feat,
    'D': empty_feat,
    'target': target,
})



def test_duplicated_keys():
    dup_df = df.rename(columns = {"C": "A"})
    with pytest.raises(Exception, match=r"X has duplicate keys `.*`"):
        WOETransformer().fit_transform(dup_df, target)

def test_woe_transformer():
    f = WOETransformer().fit_transform(feature, target)
    assert f[451] == pytest.approx(-0.17061154127869285)

def test_woe_transformer_with_str():
    f = WOETransformer().fit_transform(str_feat, target)
    assert f[451] == pytest.approx(-0.2198594761130199)

def test_woe_transformer_with_unknown_group():
    transer = WOETransformer().fit(str_feat, target)
    res = transer.transform(['Z'], default = 'min')
    assert res[0] == pytest.approx(-0.2198594761130199)

def test_woe_transformer_frame():
    res = WOETransformer().fit_transform(df, target)
    assert res.iloc[451, 1] == pytest.approx(-0.2198594761130199)

def test_woe_transformer_select_dtypes():
    res = WOETransformer().fit_transform(df, target, select_dtypes = 'object')
    assert res.loc[451, 'A'] == 3

def test_woe_transformer_exclude():
    res = WOETransformer().fit_transform(df, target, exclude = 'A')
    assert res.loc[451, 'A'] == 3

def test_woe_transformer_export_single():
    transer = WOETransformer().fit(feature, target)
    t = transer.export()
    assert t[transer._default_name][5] == pytest.approx(0.3938235330926786)

def test_woe_transformer_export():
    transer = WOETransformer().fit(df, target)
    t = transer.export()
    assert t['C'][1] == 0

def test_woe_transformer_load():
    rules = {
        'A': {
            1: 0.1,
            2: 0.2,
            3: 0.3,
        }
    }

    transer = WOETransformer().load(rules)
    assert transer._rules['A']['woe'][1] == 0.2


def test_combiner():
    f = Combiner().fit_transform(feature, target, method = 'chi')
    assert f[451] == 3

def test_combiner_with_str():
    f = Combiner().fit_transform(str_feat, target, method = 'chi')
    assert f[451] == 0

def test_combiner_unique_feature():
    f = Combiner().fit_transform(uni_feat, target, method = 'chi')
    assert f[451] == 0

def test_combiner_frame():
    res = Combiner().fit_transform(df, target)
    assert res.iloc[404, 1] == 2

def test_combiner_select_dtypes():
    res = Combiner().fit_transform(df, target, select_dtypes = 'number')
    assert res.loc[451, 'B'] == 'G'

def test_combiner_exclude():
    res = Combiner().fit_transform(df, target, exclude = 'B')
    assert res.loc[451, 'B'] == 'G'

def test_combiner_labels():
    combiner = Combiner().fit(df, target)
    res = combiner.transform(df, labels = True)
    assert res.loc[451, 'A'] == '03.[3 ~ 4)'

def test_combiner_single_feature():
    combiner = Combiner().fit(df['A'], method = 'step', n_bins = 5)
    res = combiner.transform(df['A'])
    assert res[451] == 1

def test_combiner_export():
    combiner = Combiner().fit(df, target, method = 'chi', n_bins = 4)
    bins = combiner.export()
    assert isinstance(bins['B'][0], list)

def test_combiner_update():
    combiner = Combiner().fit(df, target, method = 'chi', n_bins = 4)
    combiner.update({'A': [1,2,3,4,5,6]})
    bins = combiner.export()
    assert len(bins['A']) == 6

def test_combiner_step():
    combiner = Combiner().fit(df['A'], method = 'step', n_bins = 4)
    bins = combiner.export()
    assert bins['A'][1] == 4.5

def test_combiner_target_in_frame():
    combiner = Combiner().fit(df, 'target', n_bins = 4)
    bins = combiner.export()
    assert bins['A'][1] == 6

def test_combiner_target_in_frame_kwargs():
    combiner = Combiner().fit(df, y = 'target', n_bins = 4)
    bins = combiner.export()
    assert bins['A'][1] == 6

def test_combiner_empty_separate():
    combiner = Combiner()
    bins = combiner.fit_transform(df, 'target', n_bins = 4, empty_separate = True)
    mask = pd.isna(df['D'])
    assert (bins['D'][~mask] != 4).all()

def test_combiner_labels_with_empty():
    combiner = Combiner().fit(df, 'target', n_bins = 4, empty_separate = True)
    res = combiner.transform(df, labels = True)
    assert res.loc[2, 'D'] == '04.nan'

def test_gbdt_transformer():
    np.random.seed(1)

    df = pd.DataFrame({
        'A': np.random.rand(500),
        'B': np.random.randint(10, size = 500),
    })
    f = GBDTTransformer().fit_transform(df, target, n_estimators = 10, max_depth = 2)
    assert f.shape == (500, 40)



import sys
import argparse
import pandas as pd

def func(args):
    """detect csv data

    Examples:

        toad tree -i xxx.csv
    """
    import toad
    from .tree import split_data, dtree
    args = vars(args)

    # remove func attribute
    args.pop('func')
    
    input = args.pop('input')
    target = args.pop('target')
    include = args.pop('include')
    exclude = args.pop('exclude')

    sys.stdout.write('reading data....\n')
    data = pd.read_csv(input)

    X, *tars = split_data(data, target = target)

    if include is not None:
        X = X[include]

    if exclude is not None:
        X = X.drop(columns = exclude)

    X = toad.utils.get_dummies(X)


    for t in tars:
        sys.stdout.write('analyse '+ t.name +' ...\n')
        dtree(X, t, **args)


ARGS = {
    'info': {
        'name': 'tree',
        'description': 'analyse bad rate from a csv file',
    },
    'defaults': {
        'func': func,
    },
    'args': [
        {
            'flag': ('-i', '--input'),
            'type': argparse.FileType('r', encoding='utf-8'),
            'help': 'the csv file which will be analysed',
            'required': True,
        },
        {
            'flag': ('-t', '--target'),
            'nargs': '+',
            'help': 'the target(s) will be analysed',
            'default': 'target',
        },
        {
            'flag': ('-c', '--criterion'),
            'type': str,
            'help': 'criterion to measure the quality of a split. Support "gini" (default), "entropy"',
            'default': 'gini',
        },
        {
            'flag': ('-d', '--depth'),
            'type': int,
            'help': 'the maximum depth of the tree',
            'default': None,
        },
        {
            'flag': ('-s', '--sample'),
            'type': float,
            'help': 'minimum number of sample in each node',
            'default': 0.01,
        },
        {
            'flag': ('-r', '--ratio'),
            'type': float,
            'help': 'threshold of ratio that will be highlighted',
            'default': 0.15,
        },
        {
            'flag': ('--exclude',),
            'nargs': '+',
            'help': 'feature names that will not use to analyse',
            'default': None,
        },
        {
            'flag': ('--include',),
            'nargs': '+',
            'help': 'feature names that will be used to analyse',
            'default': None,
        },
    ]
}



import numpy as np
import pandas as pd

from .impute import impute


ab = np.array(list('ABCDEFG'))
int_feat = np.random.randint(10, size = 500)
float_feat = np.random.rand(500)
str_feat = ab[np.random.choice(7, 500)]
uni_feat = np.ones(500)
# empty_feat = np.full(500, np.nan)

target = np.random.randint(2, size = 500)

df = pd.DataFrame({
    'A': int_feat,
    'B': str_feat,
    'C': uni_feat,
    'D': float_feat,
    # 'E': empty_feat,
})

mask = np.random.choice([True, False], size = 500 * 4, p = [0.95, 0.05]).reshape(500, 4)
df = df.where(mask, np.nan)


def test_impute_with_number():
    res = impute(df.drop(columns = 'B'))

    assert res.isna().sum().sum() == 0


def test_impute_with_str():
    res = impute(df)

    assert res.isna().sum().sum() == 0











import numpy as np
import pandas as pd


class Partition:
    def partition(self, data):
        """partition data

        Args:
            data (DataFrame): dataframe
        
        Returns:
            iterator -> ndarray[bool]: mask of partition data
            iterator -> str: suffix string of current partition
        """
        yield np.ones(len(data)).astype(bool), ''



class TimePartition(Partition):
    """partition data by time delta

    Args:
        base (str): column name of base time
        filter (str): column name of target time to be compared
        times (list): list of time delta`
    
    Example:

    >>> TimePartition('apply_time', 'query_time', ['30d', '90d', 'all'])

    """
    def __init__(self, base, filter, times):
        self.base = base
        self.filter = filter
        self.times = times
    

    def partition(self, data):
        base = pd.to_datetime(data[self.base])
        filter = pd.to_datetime(data[self.filter])

        for t in self.times:
            if t != 'all':
                delta = pd.Timedelta(t)
                mask = filter > (base - delta)
            else:
                mask = np.ones(len(filter)).astype(bool)
            
            yield mask, '_' + t


class ValuePartition(Partition):
    """partition data by column values

    Args:
        column (str): column name which will be used as partition
    
    Example:

    >>> ValuePartition('status')
    
    """
    def __init__(self, column):
        self.column = column
    

    def partition(self, data):
        data = data[self.column]
        unique = data.unique()

        for u in unique:
            if pd.isna(u):
                mask = data.isna()
            else:
                mask = (data == u)
            
            yield mask, '_' + str(u)


import seaborn as sns
from .utils import (
    get_axes,
    tadpole_axes,
    FIG_SIZE,
)

class Tadpole:
    def __getattr__(self, name):
        t = getattr(sns, name)
        if callable(t):
            return self.wrapsns(t)

        return t

    def wrapsns(self, f):
        @tadpole_axes
        def wrapper(*args, figure_size = FIG_SIZE, **kwargs):
            kw = kwargs.copy()
            if 'ax' not in kw:
                kw['ax'] = get_axes(size = figure_size)

            try:
                return f(*args, **kw)
            except:
                return f(*args, **kwargs)

        return wrapper



import os
import seaborn as sns
from functools import wraps
import matplotlib.pyplot as plt
from matplotlib.axes import Axes
from matplotlib.font_manager import FontProperties

sns.set_palette('muted')

CURRENT_PATH = os.path.abspath(os.path.dirname(__file__))
FONT_FILE = 'NotoSansCJKsc-Regular.otf'
FONTS_PATH = os.path.join(CURRENT_PATH, 'fonts', FONT_FILE)
myfont = FontProperties(fname = os.path.abspath(FONTS_PATH))
sns.set(font = myfont.get_family())

HEATMAP_CMAP = sns.diverging_palette(240, 10, as_cmap = True)
MAX_STYLE = 6
FIG_SIZE = (12, 6)

def get_axes(size = FIG_SIZE):
    _, ax = plt.subplots(figsize = size)
    return ax

def reset_legend(axes):
    if axes.get_legend() is not None:
        axes.legend(
            loc='center left',
            bbox_to_anchor=(1, 0.5),
            framealpha = 0,
            prop = myfont,
        )

    return axes

def reset_ticklabels(axes):
    labels = []
    if axes.get_xticklabels():
        labels += axes.get_xticklabels()

    if axes.get_yticklabels():
        labels += axes.get_yticklabels()

    for label in labels:
        label.set_fontproperties(myfont)

    return axes

def reset_xticks(axes):
    for label in axes.get_xticklabels():
        label.set_ha('left')
        label.set_rotation(-25)
    
    return axes


def reset_title(axes):
    title = axes.get_title()
    
    if title:
        axes.set_title(title, fontproperties = myfont)
    
    return axes


def reset_xylabels(axes):
    y_label = axes.get_ylabel()
    if y_label:
        axes.set_ylabel(y_label, fontproperties = myfont)
    
    x_label = axes.get_xlabel()
    if x_label:
        axes.set_xlabel(x_label, fontproperties = myfont)
    
    return axes


def reset_ylim(axes):
    # for axes and twins
    for ax in axes.figure.axes:
        if ax.bbox.bounds == axes.bbox.bounds:
            bottom, top = ax.get_ylim()
            top += (top - bottom) * 0.1
            ax.set_ylim(bottom, top)

    return axes


def fix_axes(axes):
    if not isinstance(axes, Axes):
        return axes

    functions = [reset_title, reset_xylabels, reset_ticklabels, reset_legend, reset_xticks]

    for func in functions:
        func(axes)
    return axes

def tadpole_axes(fn):
    @wraps(fn)
    def func(*args, **kwargs):
        res = fn(*args, **kwargs)

        if not isinstance(res, tuple):
            return fix_axes(res)

        r = tuple()
        for i in res:
            r += (fix_axes(i),)

        return r

    return func



def annotate(ax, x, y, space = 5, format = ".2f"):
    """
    """
    va = 'bottom'

    if y < 0:
        space *= -1
        va = 'top'

    ax.annotate(
        ("{:"+ format +"}").format(y),
        (x, y),
        xytext = (0, space),
        textcoords = "offset points",
        ha = 'center',
        va = va,
    )



def add_bar_annotate(ax, **kwargs):
    """
    """
    for rect in ax.patches:
        y_value = rect.get_height()
        x_value = rect.get_x() + rect.get_width() / 2

        annotate(ax, x_value, y_value, **kwargs)

    return ax


def add_line_annotate(ax, **kwargs):
    """
    """
    for line in ax.lines:
        points = line.get_xydata()

        for point in points:
            annotate(ax, point[0], point[1], **kwargs)

    return ax


def add_annotate(ax, **kwargs):
    if len(ax.lines) > 0:
        add_line_annotate(ax, **kwargs)

    if len(ax.patches) > 0:
        add_bar_annotate(ax, **kwargs)

    return ax


def add_text(ax, text, loc = 'top left', offset = (0.01, 0.04)):
    x_min, x_max = ax.get_xlim()
    y_min, y_max = ax.get_ylim()

    x_offset = (x_max - x_min) * offset[0]
    y_offset = (y_max - y_min) * offset[1]

    if loc == 'top left':
        loc = (x_min + x_offset, y_max - y_offset)
    elif loc == 'top right':
        loc = (x_max - x_offset, y_max - y_offset)
    elif loc == 'bottom left':
        loc = (x_min + x_offset, y_min + y_offset)
    elif loc == 'bottom right':
        loc = (x_max - x_offset, y_min + y_offset)

    ax.text(*loc, text, fontsize = 'x-large')

    return ax



import torch
import numpy as np
from .history import History

def test_history_log():
    history = History()

    for i in range(10):
        history.log('tensor', torch.rand(3, 5))
    
    assert history['tensor'].shape == (30, 5)





import pytest
import numpy as np
import pandas as pd

from .plot import (
    badrate_plot,
    corr_plot,
    proportion_plot,
    roc_plot,
    bin_plot,
)

np.random.seed(1)

LENGTH = 500

A = np.random.rand(LENGTH)
A[np.random.choice(LENGTH, 20, replace = False)] = np.nan

B = np.random.randint(100, size = LENGTH)
C = A + np.random.normal(0, 0.2, LENGTH)
D = A + np.random.normal(0, 0.1, LENGTH)

E = np.random.rand(LENGTH)
E[np.random.choice(LENGTH, 480, replace = False)] = np.nan

F = B + np.random.normal(0, 10, LENGTH)

target = np.random.randint(2, size = LENGTH)

frame = pd.DataFrame({
    'A': A,
    'B': B,
    'C': C,
    'D': D,
    'E': E,
    'F': F,
})

frame['target'] = target


def test_badrate_plot():
    g = badrate_plot(
        frame,
        x = 'A',
        target = 'target',
        return_counts = True,
        return_proportion = True,
    )

def test_badrate_plot_y_axis():
    g = badrate_plot(
        frame,
        x = 'A',
        target = 'target',
    )
    bottom, _ = g.get_ylim()
    assert bottom == 0

def test_corr_plot():
    g = corr_plot(frame)


def test_proportion_plot():
    g = proportion_plot(x = frame['target'])


def test_roc_plot():
    g = roc_plot(frame['B'], frame['target'])


def test_bin_plot():
    g = bin_plot(frame, x = 'B', target = 'target')


import re
import numpy as np
from copy import deepcopy
from .decorator import save_to_json, load_from_json


DEFAULT_NAME = '_feature_default_name_'


class RulesMixin:
    _rules = {}

    def _parse_rule(self, rule):
        return rule
    
    def _format_rule(self, rule):
        return rule
    
    def default_rule(self):
        if len(self._rules) == 1:
            # return the only rule as default
            return next(iter(self._rules.values()))
        
        if self._default_name not in self._rules:
            raise Exception('can not get default rule')

        return self._rules[self._default_name]
    
    @property
    def _default_name(self):
        return DEFAULT_NAME

    @property
    def rules(self):
        return self._rules
    
    @rules.setter
    def rules(self, value):
        self._rules = value
    

    @load_from_json(is_class = True, require_first = True)
    def load(self, rules, update = False, **kwargs):
        """load rules from dict or json file

        Args:
            rules (dict): dictionary of rules
            from_json (str|IOBase): json file of rules
            update (bool): if need to use updating instead of replacing rules
        """
        rules = deepcopy(rules)

        if not isinstance(rules, dict):
            rules = {
                self._default_name: rules,
            }
        
        for key in rules:
            rules[key] = self._parse_rule(rules[key], **kwargs)
        
        if update:
            self._rules.update(rules)
        else:
            self._rules = rules
        
        if hasattr(self, 'after_load'):
            self.after_load(rules)
        
        return self
    
    @save_to_json(is_class = True)
    def export(self, **kwargs):
        """export rules to dict or a json file

        Args:
            to_json (str|IOBase): json file to save rules
        
        Returns:
            dict: dictionary of rules
        """
        res = {}
        for key in self._rules:
            res[key] = self._format_rule(self._rules[key], **kwargs)
        
        if hasattr(self, 'after_export'):
            res = self.after_export(res, **kwargs)
        
        return res
    
    def update(self, *args, **kwargs):
        """update rules

        Args:
            rules (dict): dictionary of rules
            from_json (str|IOBase): json file of rules
        """
        return self.load(*args, update = True, **kwargs)
    

    def __len__(self):
        return len(self._rules.keys())
    
    def __contains__(self, key):
        return key in self._rules
    
    def __getitem__(self, key):
        return self._rules[key]
    
    def __setitem__(self, key, value):
        self._rules[key] = value

    def __iter__(self):
        return iter(self._rules)




RE_NUM = r'-?\d+(.\d+)?'
RE_SEP = r'[~-]'
RE_BEGIN = r'(-inf|{num})'.format(num = RE_NUM)
RE_END = r'(inf|{num})'.format(num = RE_NUM)
RE_RANGE = r'\[{begin}\s*{sep}\s*{end}\)'.format(
    begin = RE_BEGIN,
    end = RE_END,
    sep = RE_SEP,
)





class BinsMixin:
    EMPTY_BIN = -1
    ELSE_GROUP = 'else'
    NUMBER_EXP = re.compile(RE_RANGE)

    @classmethod
    def parse_bins(self, bins):
        """parse labeled bins to array
        """
        if self._is_numeric(bins):
            return self._numeric_parser(bins)
        
        l = list()

        for item in bins:
            if item == self.ELSE_GROUP:
                l.append(item)
            else:
                l.append(item.split(','))

        return np.array(l, dtype = object)


    @classmethod
    def format_bins(self, bins, index = False, ellipsis = None):
        """format bins to label

        Args:
            bins (ndarray): bins to format
            index (bool): if need index prefix
            ellipsis (int): max length threshold that labels will not be ellipsis, `None` for skipping ellipsis
        
        Returns:
            ndarray: array of labels
        """
        l = list()

        if np.issubdtype(bins.dtype, np.number):
            has_empty = len(bins) > 0 and np.isnan(bins[-1])
            
            if has_empty:
                bins = bins[:-1]
            
            sp_l = [-np.inf] + bins.tolist() + [np.inf]
            for i in range(len(sp_l) - 1):
                l.append('['+str(sp_l[i])+' ~ '+str(sp_l[i+1])+')')
            
            if has_empty:
                l.append('nan')
        else:
            for keys in bins:
                if isinstance(keys, str) and keys == self.ELSE_GROUP:
                    l.append(keys)
                else:
                    label = ','.join(keys)
                    if ellipsis is not None:
                        label = label[:ellipsis] + '..' if len(label) > ellipsis else label
                    l.append(label)

        if index:
            l = ["{:02}.{}".format(ix, lab) for ix, lab in enumerate(l)]

        return np.array(l)
    

    @classmethod
    def _is_numeric(self, bins):
        m = self.NUMBER_EXP.match(bins[0])

        return m is not None
    
    @classmethod
    def _numeric_parser(self, bins):
        l = list()

        for item in bins:

            if item == 'nan':
                l.append(np.nan)
                continue
            
            m = self.NUMBER_EXP.match(item)
            split = m.group(3)

            if split == 'inf':
                # split = np.inf
                continue
            
            split = float(split)

            l.append(split)
        
        return np.array(l)



import pytest
import numpy as np
from .mixin import RulesMixin, BinsMixin

np.random.seed(1)

class RulesObject(RulesMixin):
    def _parse_rule(self, rule):
        return {
            'rule': rule
        }


    def _format_rule(self, rule):
        return 'rule -> %s' % rule['rule']


rules = {'A': 'rule_A'}

def test_rule_parse():
    r = RulesObject().load(rules)
    assert r.rules['A']['rule'] == 'rule_A'

def test_rule_format():
    r = RulesObject().load(rules)
    assert r.export()['A'] == 'rule -> rule_A'

def test_save_update():
    r = RulesObject().load(rules)
    r.update({'A': 'update_A'})
    assert r.rules['A']['rule'] == 'update_A'

def test_format_bins():
    obj = BinsMixin()
    formated = obj.format_bins(np.array([2,4,6]))
    expect = ['[-inf ~ 2)', '[2 ~ 4)', '[4 ~ 6)', '[6 ~ inf)']
    assert all([a == b for a, b in zip(formated, expect)])

def test_format_bins_with_index():
    obj = BinsMixin()
    formated = obj.format_bins(np.array([2,4,6]), index = True)
    assert '01.[2 ~ 4)' in formated

def test_format_bins_with_ellipsis():
    obj = BinsMixin()
    formated = obj.format_bins(np.array([['A', 'B', 'C'], ['D', 'E']], dtype = object), ellipsis = 3)
    assert formated[0] == 'A,B..' and formated[1] == 'D,E'



import pytest
import numpy as np
import pandas as pd

from .func import (
    np_unique,
    fillna,
    clip,
    diff_time_frame,
    bin_to_number,
    generate_target,
    generate_str,
    get_dummies,
    feature_splits,
)

np.random.seed(1)
feature = np.random.rand(500)
target = np.random.randint(2, size = 500)



def test_fillna():
    res = fillna(np.array([1, 2, 3, np.nan, 4, 5]))
    assert res[3] == -1


def test_np_unique():
    res = np_unique(np.array([np.nan, np.nan, np.nan]))
    assert len(res) == 1


def test_clip():
    res1 = clip(feature, quantile = (.05, .95))
    res2 = clip(feature, quantile = 0.05)
    assert np.testing.assert_array_equal(res1, res2) is None


def test_feature_splits():
    value = feature_splits(feature, target)
    assert len(value) == 243


def test_diff_time_frame():
    time_data = [
        {
            'base': '2018-01',
            'time1': '2018-04',
            'time2': '2018-04-02',
        },
        {
            'base': '2018-01',
            'time1': '2018-05',
            'time2': '2018-04-05',
        },
        {
            'base': '2018-02',
            'time1': '2018-04',
            'time2': '2018-04-10',
        },
    ]

    frame = pd.DataFrame(time_data)
    res = diff_time_frame(frame['base'], frame[['time1', 'time2']], format='%Y-%m-%d')
    assert res.iloc[0, 1] == 91

def test_bin_to_number():
    s = pd.Series([
        '1',
        '1-100',
        '-',
        '100-200',
        np.nan,
        '200-300',
        '300',
        '100-200',
        '>500',
    ])

    res = s.apply(bin_to_number())
    assert res[3] == 150

def test_bin_to_number_for_frame():
    df = pd.DataFrame([
        {
            'area_1': '100-200',
            'area_2': '150~200',
        },
        {
            'area_1': '300-400',
            'area_2': '200~250',
        },
        {
            'area_1': '200-300',
            'area_2': '450~500',
        },
        {
            'area_1': '100-200',
            'area_2': '250~300',
        },
    ])

    res = df.applymap(bin_to_number())
    assert res.loc[1, 'area_2'] == 225

def test_generate_target():
    t = generate_target(len(feature), rate = 0.3, weight = feature)
    rate = t.sum() / len(t)
    assert rate == 0.3

@pytest.fixture
def test_generate_str():
    s = generate_str(size = 8)
    assert s == 'EPL5MTQK'

def test_get_dummies_binary():
    ab = np.array(list('ABCDEFG'))
    df = pd.DataFrame({
        'binary': ab[np.random.choice(2, 500)],
        'multiple': ab[np.random.choice(5, 500)],
    })
    data = get_dummies(df, binary_drop = True)
    
    assert 'binary_A' not in data.columns



import numpy as np
import pandas as pd
from scipy import stats

from .stats import IV, VIF
from .metrics import MSE, AIC, BIC, KS, AUC
from .utils import split_target, unpack_tuple, to_ndarray


INTERCEPT_COLS = 'intercept'


class StatsModel:
    def __init__(self, estimator = 'ols', criterion = 'aic', intercept = False):
        if isinstance(estimator, str):
            Est = self.get_estimator(estimator)
            estimator = Est(fit_intercept = intercept,)

        self.estimator = estimator
        self.intercept = intercept
        self.criterion = criterion


    def get_estimator(self, name):
        from sklearn.linear_model import (
            LinearRegression,
            LogisticRegression,
            Lasso,
            Ridge,
        )

        ests = {
            'ols': LinearRegression,
            'lr': LogisticRegression,
            'lasso': Lasso,
            'ridge': Ridge,
        }

        if name in ests:
            return ests[name]

        raise Exception('estimator {name} is not supported'.format(name = name))



    def stats(self, X, y):
        """
        """
        X = X.copy()

        if isinstance(X, pd.Series):
            X = X.to_frame()

        self.estimator.fit(X, y)

        if hasattr(self.estimator, 'predict_proba'):
            pre = self.estimator.predict_proba(X)[:, 1]
        else:
            pre = self.estimator.predict(X)

        coef = self.estimator.coef_.reshape(-1)

        if self.intercept:
            coef = np.append(coef, self.estimator.intercept_)
            X[INTERCEPT_COLS] = np.ones(X.shape[0])

        n, k = X.shape

        t_value = self.t_value(pre, y, X, coef)
        p_value = self.p_value(t_value, n)
        c = self.get_criterion(pre, y, k)

        return {
            't_value': pd.Series(t_value, index = X.columns),
            'p_value': pd.Series(p_value, index = X.columns),
            'criterion': c
        }

    def get_criterion(self, pre, y, k):
        if self.criterion == 'aic':
            llf = self.loglikelihood(pre, y, k)
            return AIC(pre, y, k, llf = llf)

        if self.criterion == 'bic':
            llf = self.loglikelihood(pre, y, k)
            return BIC(pre, y, k, llf = llf)

        if self.criterion == 'ks':
            return KS(pre, y)

        if self.criterion == 'auc':
            return AUC(pre, y)

    def t_value(self, pre, y, X, coef):
        n, k = X.shape
        mse = sum((y - pre) ** 2) / float(n - k)
        nx = np.dot(X.T, X)

        if np.linalg.det(nx) == 0:
            return np.nan

        std_e = np.sqrt(mse * (np.linalg.inv(nx).diagonal()))
        return coef / std_e

    def p_value(self, t, n):
        return stats.t.sf(np.abs(t), n - 1) * 2

    def loglikelihood(self, pre, y, k):
        n = len(y)
        mse = MSE(pre, y)
        return (-n / 2) * np.log(2 * np.pi * mse * np.e)


def stepwise(frame, target = 'target', estimator = 'ols', direction = 'both', criterion = 'aic',
            p_enter = 0.01, p_remove = 0.01, p_value_enter = 0.2, intercept = False,
            max_iter = None, return_drop = False, exclude = None):
    """stepwise to select features

    Args:
        frame (DataFrame): dataframe that will be use to select
        target (str): target name in frame
        estimator (str): model to use for stats
        direction (str): direction of stepwise, support 'forward', 'backward' and 'both', suggest 'both'
        criterion (str): criterion to statistic model, support 'aic', 'bic'
        p_enter (float): threshold that will be used in 'forward' and 'both' to keep features
        p_remove (float): threshold that will be used in 'backward' to remove features
        intercept (bool): if have intercept
        p_value_enter (float): threshold that will be used in 'both' to remove features
        max_iter (int): maximum number of iterate
        return_drop (bool): if need to return features' name who has been dropped
        exclude (array-like): list of feature names that will not be dropped

    Returns:
        DataFrame: selected dataframe
        array: list of feature names that has been dropped
    """
    df, y = split_target(frame, target)

    if exclude is not None:
        df = df.drop(columns = exclude)

    drop_list = []
    remaining = df.columns.tolist()

    selected = []

    sm = StatsModel(estimator = estimator, criterion = criterion, intercept = intercept)

    order = -1 if criterion in ['aic', 'bic'] else 1

    best_score = -np.inf * order

    iter = -1
    while remaining:
        iter += 1
        if max_iter and iter > max_iter:
            break

        l = len(remaining)
        test_score = np.zeros(l)
        test_res = np.empty(l, dtype = np.object)

        if direction == 'backward':
            for i in range(l):
                test_res[i] = sm.stats(
                    df[ remaining[:i] + remaining[i+1:] ],
                    y,
                )
                test_score[i] = test_res[i]['criterion']

            curr_ix = np.argmax(test_score * order)
            curr_score = test_score[curr_ix]

            if (curr_score - best_score) * order < p_remove:
                break

            name = remaining.pop(curr_ix)
            drop_list.append(name)

            best_score = curr_score

        # forward and both
        else:
            for i in range(l):
                test_res[i] = sm.stats(
                    df[ selected + [remaining[i]] ],
                    y,
                )
                test_score[i] = test_res[i]['criterion']

            curr_ix = np.argmax(test_score * order)
            curr_score = test_score[curr_ix]

            name = remaining.pop(curr_ix)
            if (curr_score - best_score) * order < p_enter:
                drop_list.append(name)

                # early stop
                if selected:
                    drop_list += remaining
                    break

                continue

            selected.append(name)
            best_score = curr_score

            if direction == 'both':
                p_values = test_res[curr_ix]['p_value']
                drop_names = p_values[p_values > p_value_enter].index

                for name in drop_names:
                    selected.remove(name)
                    drop_list.append(name)

    r = frame.drop(columns = drop_list)

    res = (r,)
    if return_drop:
        res += (drop_list,)

    return unpack_tuple(res)


def drop_empty(frame, threshold = 0.9, nan = None, return_drop = False,
            exclude = None):
    """drop columns by empty

    Args:
        frame (DataFrame): dataframe that will be used
        threshold (number): drop the features whose empty num is greater than threshold. if threshold is float, it will be use as percentage
        nan (any): values will be look like empty
        return_drop (bool): if need to return features' name who has been dropped
        exclude (array-like): list of feature names that will not be dropped

    Returns:
        DataFrame: selected dataframe
        array: list of feature names that has been dropped
    """
    cols = frame.columns.copy()

    if exclude is not None:
        cols = cols.drop(exclude)

    if threshold < 1:
        threshold = len(frame) * threshold

    drop_list = []
    for col in cols:
        series = frame[col]
        
        if nan is not None:
            series = series.replace(nan, np.nan)
        
        n = series.isnull().sum()
        if n > threshold:
            drop_list.append(col)

    r = frame.drop(columns = drop_list)

    res = (r,)
    if return_drop:
        res += (np.array(drop_list),)

    return unpack_tuple(res)


def drop_var(frame, threshold = 0, return_drop = False, exclude = None):
    """drop columns by variance

    Args:
        frame (DataFrame): dataframe that will be used
        threshold (float): drop features whose variance is less than threshold
        return_drop (bool): if need to return features' name who has been dropped
        exclude (array-like): list of feature names that will not be dropped

    Returns:
        DataFrame: selected dataframe
        array: list of feature names that has been dropped
    """
    df = frame.copy()

    if exclude is not None:
        df = df.drop(columns = exclude)

    # numeric features only
    df = df.select_dtypes(include = 'number')

    variances = np.var(df, axis = 0)
    drop_list = df.columns[variances <= threshold]

    r = frame.drop(columns = drop_list)

    res = (r,)
    if return_drop:
        res += (drop_list)

    return unpack_tuple(res)


def drop_corr(frame, target = None, threshold = 0.7, by = 'IV',
            return_drop = False, exclude = None):
    """drop columns by correlation

    Args:
        frame (DataFrame): dataframe that will be used
        target (str): target name in dataframe
        threshold (float): drop features that has the smallest weight in each groups whose correlation is greater than threshold
        by (array-like): weight of features that will be used to drop the features
        return_drop (bool): if need to return features' name who has been dropped
        exclude (array-like): list of feature names that will not be dropped

    Returns:
        DataFrame: selected dataframe
        array: list of feature names that has been dropped
    """
    if not isinstance(by, (str, pd.Series)):
        by = pd.Series(by, index = frame.columns)

    cols = frame.columns.copy()

    if exclude is not None:
        exclude = exclude if isinstance(exclude, (list, np.ndarray)) else [exclude]
        cols = cols.drop(exclude)


    f, t = split_target(frame[cols], target)

    corr = f.corr().abs()

    drops = []

    # get position who's corr greater than threshold
    ix, cn = np.where(np.triu(corr.values, 1) > threshold)

    # if has position
    if len(ix):
        # get the graph of relationship
        graph = np.hstack([ix.reshape((-1, 1)), cn.reshape((-1, 1))])

        uni, counts = np.unique(graph, return_counts = True)

        # calc weights for nodes
        weights = np.zeros(len(corr.index))


        if isinstance(by, pd.Series):
            weights = by[corr.index].values
        elif by.upper() == 'IV':
            for ix in uni:
                weights[ix] = IV(frame[corr.index[ix]], target = t)


        while(True):
            # TODO deal with circle

            # get nodes with the most relationship
            nodes = uni[np.argwhere(counts == np.amax(counts))].flatten()

            # get node who has the min weights
            n = nodes[np.argsort(weights[nodes])[0]]

            # get nodes of 1 degree relationship of n
            i, c = np.where(graph == n)
            pairs = graph[(i, 1-c)]

            # if sum of 1 degree nodes greater than n
            # then delete n self
            # else delete all 1 degree nodes
            if weights[pairs].sum() > weights[n]:
                dro = [n]
            else:
                dro = pairs.tolist()

            # add nodes to drops list
            drops += dro

            # delete nodes from graph
            di, _ = np.where(np.isin(graph, dro))
            graph = np.delete(graph, di, axis = 0)

            # if graph is empty
            if len(graph) <= 0:
                break

            # update nodes and counts
            uni, counts = np.unique(graph, return_counts = True)


    drop_list = corr.index[drops].values
    r = frame.drop(columns = drop_list)

    res = (r,)
    if return_drop:
        res += (drop_list,)

    return unpack_tuple(res)


def drop_iv(frame, target = 'target', threshold = 0.02, return_drop = False,
            return_iv = False, exclude = None):
    """drop columns by IV

    Args:
        frame (DataFrame): dataframe that will be used
        target (str): target name in dataframe
        threshold (float): drop the features whose IV is less than threshold
        return_drop (bool): if need to return features' name who has been dropped
        return_iv (bool): if need to return features' IV
        exclude (array-like): list of feature names that will not be dropped

    Returns:
        DataFrame: selected dataframe
        array: list of feature names that has been dropped
        Series: list of features' IV
    """
    cols = frame.columns.copy()

    if exclude is not None:
        cols = cols.drop(exclude)

    f, t = split_target(frame[cols], target)

    l = len(f.columns)
    iv = np.zeros(l)

    for i in range(l):
        iv[i] = IV(f[f.columns[i]], target = t)

    drop_ix = np.where(iv < threshold)

    drop_list = f.columns[drop_ix].values
    df = frame.drop(columns = drop_list)

    res = (df,)
    if return_drop:
        res += (drop_list,)

    if return_iv:
        res += (pd.Series(iv, index = f.columns),)

    return unpack_tuple(res)


def drop_vif(frame, threshold = 3, return_drop = False, exclude = None):
    """variance inflation factor

    Args:
        frame (DataFrame)
        threshold (float): drop features until all vif is less than threshold
        return_drop (bool): if need to return features' name who has been dropped
        exclude (array-like): list of feature names that will not be dropped

    Returns:
        DataFrame: selected dataframe
        array: list of feature names that has been dropped
    """
    cols = frame.columns.copy()

    if exclude is not None:
        cols = cols.drop(exclude)

    drop_list = []
    while(True):
        vif = VIF(frame[cols])

        ix = vif.idxmax()
        max = vif[ix]

        if max < threshold:
            break

        cols = cols.drop(ix)
        drop_list.append(ix)


    r = frame.drop(columns = drop_list)

    res = (r,)
    if return_drop:
        res += (drop_list,)

    return unpack_tuple(res)


def select(frame, target = 'target', empty = 0.9, iv = 0.02, corr = 0.7,
            return_drop = False, exclude = None):
    """select features by rate of empty, iv and correlation

    Args:
        frame (DataFrame)
        target (str): target's name in dataframe
        empty (number): drop the features which empty num is greater than threshold. if threshold is float, it will be use as percentage
        iv (float): drop the features whose IV is less than threshold
        corr (float): drop features that has the smallest IV in each groups which correlation is greater than threshold
        return_drop (bool): if need to return features' name who has been dropped
        exclude (array-like): list of feature name that will not be dropped

    Returns:
        DataFrame: selected dataframe
        dict: list of dropped feature names in each step
    """
    empty_drop = iv_drop = corr_drop = None

    if empty is not False:
        frame, empty_drop = drop_empty(frame, threshold = empty, return_drop = True, exclude = exclude)

    if iv is not False:
        frame, iv_drop, iv_list = drop_iv(frame, target = target, threshold = iv, return_drop = True, return_iv = True, exclude = exclude)

    if corr is not False:
        weights = 'IV'

        if iv is not False:
            weights = iv_list

        frame, corr_drop = drop_corr(frame, target = target, threshold = corr, by = weights, return_drop = True, exclude = exclude)

    res = (frame,)
    if return_drop:
        d = {
            'empty': empty_drop,
            'iv': iv_drop,
            'corr': corr_drop,
        }
        res += (d,)

    return unpack_tuple(res)



import pytest
import numpy as np
import pandas as pd

from .stats import IV, WOE, gini, gini_cond, entropy_cond, quality, _IV, VIF


np.random.seed(1)

feature = np.random.rand(500)
target = np.random.randint(2, size = 500)
A = np.random.randint(100, size = 500)
B = np.random.randint(100, size = 500)
mask = np.random.randint(8, size = 500)

df = pd.DataFrame({
    'feature': feature,
    'target': target,
    'A': A,
    'B': B,
})


def test_woe():
    value = WOE(0.2, 0.3)
    assert value == pytest.approx(-0.4054651081081643)

def test_iv_priv():
    value, _ = _IV(df['feature'], df['target'])
    assert value == pytest.approx(0.010385942643745403)

def test_iv():
    value = IV(df['feature'], df['target'], n_bins = 10, method = 'dt')
    assert value == pytest.approx(0.2735917707743619)

def test_iv_return_sub():
    _, sub = IV(mask, df['target'], return_sub = True, n_bins = 10, method = 'dt')
    assert len(sub) == 8
    assert sub[4] == pytest.approx(0.006449386778057019)

def test_iv_frame():
    res = IV(df, 'target', n_bins = 10, method = 'chi')
    assert res.loc[0, 'A'] == pytest.approx(0.226363832867123)

def test_gini():
    value = gini(df['target'])
    assert value == 0.499352

def test_gini_cond():
    value = gini_cond(df['feature'], df['target'])
    assert value == pytest.approx(0.4970162601626016)

def test_entropy_cond():
    value = entropy_cond(df['feature'], df['target'])
    assert value == pytest.approx(0.6924990371522171)

def test_quality():
    result = quality(df, 'target')
    assert result.loc['feature', 'iv'] == 0.2735917707743619
    assert result.loc['A', 'gini'] == 0.49284164671885444
    assert result.loc['B', 'entropy'] == 0.6924956879070063
    assert result.loc['feature', 'unique'] == 500

def test_quality_iv_only():
    result = quality(df, 'target', iv_only = True)
    assert np.isnan(result.loc['feature', 'gini'])

def test_quality_with_merge():
    result = quality(df, 'target', n_bins = 5, method = 'chi')
    assert result.loc['feature', 'iv'] == 0.13367825777558

def test_quality_object_type_array_with_nan():
    feature = np.array([np.nan, 'A', 'B', 'C', 'D', 'E', 'F', 'G'], dtype = 'O')[mask]

    df = pd.DataFrame({
        'feature': feature,
        'target': target,
    })
    result = quality(df)
    assert result.loc['feature', 'iv'] == 0.016379338180530334

def test_vif():
    vif = VIF(df)
    assert vif['A'] == 2.969336442640111



import pytest
import numpy as np
import pandas as pd


from .partition import TimePartition, ValuePartition


np.random.seed(1)

ab = np.array(list('ABCDEFG'))

history = np.full(500, np.datetime64('2020-03-01')) - np.random.randint(30, 400, size = 500)
open_time = np.full(500, np.datetime64('2020-03-01')) - np.random.randint(30, size = 500)
A = ab[np.random.choice(7, 500)]
B = np.random.randint(10, size = 500).astype(float)
B[np.random.choice(500, 10)] = np.nan


df = pd.DataFrame({
    'history': history,
    'open_time': open_time,
    'A': A,
    'B': B,
})


def test_timepartition():
    tp = TimePartition('open_time', 'history', ['90d', '180d'])
    mask, suffix = next(tp.partition(df))
    assert mask.sum() == 93


def test_timepartition_all():
    tp = TimePartition('open_time', 'history', ['all'])
    mask, suffix = next(tp.partition(df))
    assert mask.sum() == 500

def test_valuepartition():
    vp = ValuePartition('A')
    mask, suffix = next(vp.partition(df))
    assert mask.sum() == 67

def test_valuepartition_with_na():
    vp = ValuePartition('B')
    s = 0
    for mask, suffix in vp.partition(df):
        s += mask.sum()
    
    assert s == 500


import numpy as np
import pandas as pd
from pandas.api.types import is_numeric_dtype
from sklearn.experimental import enable_iterative_imputer
from sklearn.impute import IterativeImputer
from sklearn.ensemble import RandomForestRegressor
from sklearn.preprocessing import LabelEncoder



def impute(df):
    imputer = Imputer(
        estimator = RandomForestRegressor(),
        random_state = 1,
    )

    return imputer.fit_transform(df)


class Imputer(IterativeImputer):
    def __init__(self, missing_values = np.nan, **kwargs):
        super().__init__(missing_values = np.nan, **kwargs)

        if not isinstance(missing_values, list):
            missing_values = [missing_values]
        
        self.missing_values_list = missing_values
        self.encoder_dict = dict()

    def _impute_one_feature(self, X_filled, mask_missing_values, feat_idx,
                            neighbor_feat_idx, **kwargs):
        
        return super()._impute_one_feature(X_filled, mask_missing_values, feat_idx, neighbor_feat_idx, **kwargs)

    def fit_transform(self, X, **kwargs):
        X, mask = self._replace_empty(X)
        X = self._fit_encode(X, mask)

        res = super().fit_transform(X, **kwargs)
        res = pd.DataFrame(res, columns = X.columns)
        return self._decode(res)

    
    def transform(self, X, **kwargs):
        X, mask = self._replace_empty(X)
        X = self._encode(X, mask)

        res = super().transform(X, **kwargs)
        res = pd.DataFrame(res, columns = X.columns)
        return self._decode(res)
    

    def _replace_empty(self, X):
        mask = X.isin(self.missing_values_list)
        X = X.where(~mask, np.nan)
        return X, mask

    def _fit_encode(self, X, mask):
        """fit encoder for object data

        Args:
            X (DataFrame)
            mask (Mask): empty mask for X
        """
        category_data = X.select_dtypes(exclude = np.number).columns
        
        for col in category_data:
            unique, X[col].loc[~mask[col]] = np.unique(X[col][~mask[col]], return_inverse = True)

            self.encoder_dict[col] = unique
        
        return X

    def _encode(self, X, mask):
        """encode object data to number

        Args:
            X (DataFrame)
            mask (Mask): empty mask for X
        """
        for col, unique in self.encoder_dict.items():
            table = dict(zip(unique, np.arange(len(unique))))
            X[col].loc[~mask[col]] = np.array([table[v] for v in X[col][~mask[col]]])
        
        return X
    
    def _decode(self, X):
        """decode object data from number to origin data

        Args:
            X (DataFrame)
            mask (Mask): empty mask for X
        """
        for col, unique in self.encoder_dict.items():
            ix = X[col].values.astype(int)
            X[col] = unique[ix]
        
        return X

    





import pytest
import numpy as np
import pandas as pd

from .selection import drop_empty, drop_var, drop_corr, drop_iv, drop_vif, select, stepwise

np.random.seed(1)

LENGTH = 500

A = np.random.rand(LENGTH)
A[np.random.choice(LENGTH, 20, replace = False)] = np.nan

B = np.random.randint(100, size = LENGTH)
C = A + np.random.normal(0, 0.2, LENGTH)
D = A + np.random.normal(0, 0.1, LENGTH)

E = np.random.rand(LENGTH)
E[np.random.choice(LENGTH, 480, replace = False)] = np.nan

F = B + np.random.normal(0, 10, LENGTH)

target = np.random.randint(2, size = LENGTH)

frame = pd.DataFrame({
    'A': A,
    'B': B,
    'C': C,
    'D': D,
    'E': E,
    'F': F,
})

frame['target'] = target


def test_drop_empty():
    df = drop_empty(frame, threshold = 0.8)
    assert 'E' not in df

def test_drop_var():
    df = drop_var(frame, threshold = 0.1)
    assert 'A' not in df

def test_drop_var_exclude():
    df = drop_var(frame, threshold = 0.1, exclude = 'A')
    assert 'A' in df

def test_drop_corr():
    df = drop_corr(frame, target = 'target')
    assert ['D', 'E', 'F', 'target'] == df.columns.tolist()

def test_drop_iv():
    df = drop_iv(frame, target = 'target', threshold = 0.25)
    assert 'B' not in df

def test_select():
    df = select(frame, target = 'target', empty = 0.8, iv = 0.2, corr = 0.7)
    assert ['D', 'F', 'target'] == df.columns.tolist()

def test_select_exclude():
    df = select(frame, target = 'target', empty = 0.8, iv = 0.2, corr = 0.7, exclude = ['A'])
    assert ['A', 'D', 'F', 'target'] == df.columns.tolist()

def test_stepwise():
    df = stepwise(frame.fillna(-1), target = 'target')
    assert ['C', 'E', 'F', 'target'] == df.columns.tolist()

def test_stepwise_backward():
    df = stepwise(frame.fillna(-1), target = 'target', direction = 'backward')
    assert ['C', 'E', 'F', 'target'] == df.columns.tolist()

def test_stepwise_forward():
    df = stepwise(frame.fillna(-1), target = 'target', direction = 'forward')
    assert ['C', 'E', 'F', 'target'] == df.columns.tolist()

def test_stepwise_exclude():
    df = stepwise(frame.fillna(-1), target = 'target', exclude = 'B')
    assert ['B', 'C', 'E', 'F', 'target'] == df.columns.tolist()

def test_stepwise_return_drop():
    df, drop_list = stepwise(frame.fillna(-1), target = 'target', return_drop = True)
    assert ['B', 'A', 'D'] == drop_list

def test_stepwise_lr():
    df = stepwise(frame.fillna(-1), target = 'target', estimator = 'lr', direction = 'forward')
    assert ['C', 'target'] == df.columns.tolist()

def test_stepwise_ks():
    df = stepwise(frame.fillna(-1), target = 'target', criterion = 'ks', direction = 'forward')
    assert ['A', 'C', 'target'] == df.columns.tolist()

def test_stepwise_zero():
    df = pd.DataFrame({
        'X': np.zeros(500),
        'Z': np.random.rand(500),
        'Y': np.random.randint(2, size = 500),
    })
    df = stepwise(df, target = 'Y')
    assert set(['Z', 'Y']) == set(df.columns.tolist())

def test_stepwise_forward_when_best_is_first():
    df = frame[['E', 'F', 'B', 'A', 'D', 'C', 'target']]
    df = stepwise(df.fillna(-1), target = 'target', direction = 'forward')
    assert ['E', 'F', 'C', 'target'] == df.columns.tolist()

def test_drop_vif():
    df = drop_vif(frame.fillna(-1), exclude = 'target')
    assert ['C', 'F', 'target'] == df.columns.tolist()



import torch
from torch import nn
import torch.nn.functional as F
from torch.utils.data import TensorDataset, DataLoader

from ..module import Module
from .trainer import Trainer
from .earlystopping import EarlyStopping


DATASET_SIZE = 20000
NUM_FEATS = 784
NUM_CLASSES = 2

X = torch.rand(DATASET_SIZE, NUM_FEATS, dtype = torch.float)
y = torch.randint(NUM_CLASSES, size = (DATASET_SIZE,), dtype = torch.long)

loader = DataLoader(
    TensorDataset(X, y),
    batch_size = 128,
    shuffle = True,
)

class TestModel(Module):
    def __init__(self, in_feats, out_feats):
        super().__init__()

        self.linear = nn.Linear(in_feats, out_feats)
    
    def forward(self, x):
        x = self.linear(x)
        return F.relu(x)
    
    def fit_step(self, batch):
        x, y = batch
        y_hat = self(x)
        return F.cross_entropy(y_hat, y)


def test_trainer():
    model = TestModel(NUM_FEATS, NUM_CLASSES)
    trainer = Trainer(model, loader)
    trainer.train(epoch = 2)
    assert len(trainer.history) == 2


def test_trainer_early_stopping():
    model = TestModel(NUM_FEATS, NUM_CLASSES)
    early_stopping = EarlyStopping(delta = -1.0, patience = 3)
    trainer = Trainer(model, loader, early_stopping = early_stopping)
    trainer.train(epoch = 200)
    assert len(trainer.history) == 4



import pytest
import numpy as np
import pandas as pd

import pyximport

pyximport.install(setup_args={"include_dirs": np.get_include()})

from .merge import merge, ChiMerge, DTMerge, QuantileMerge, StepMerge, KMeansMerge

np.random.seed(1)
feature = np.random.rand(500)
target = np.random.randint(2, size = 500)
A = np.random.randint(100, size = 500)
B = np.random.randint(3, size = 500)

df = pd.DataFrame({
    'feature': feature,
    'target': target,
    'A': A,
})



def test_chimerge():
    splits = ChiMerge(feature, target, n_bins = 10)
    assert len(splits) == 9

def test_chimerge_bins_not_enough():
    splits = ChiMerge(B, target, n_bins = 10)
    assert len(splits) == 2

def test_chimerge_bins_with_min_samples():
    splits = ChiMerge(feature, target, min_samples = 0.02)
    assert len(splits) == 10

def test_dtmerge():
    splits = DTMerge(feature, target, n_bins = 10)
    assert len(splits) == 9

def test_quantilemerge():
    splits = QuantileMerge(feature, n_bins = 10)
    assert len(splits) == 9

def test_quantilemerge_not_enough():
    splits = QuantileMerge(B, n_bins = 10)
    assert len(splits) == 2

def test_stepmerge():
    splits = StepMerge(feature, n_bins = 10)
    assert len(splits) == 9

def test_kmeansmerge():
    splits = KMeansMerge(feature, n_bins = 10)
    assert len(splits) == 9

def test_merge():
    res = merge(feature, target = target, method = 'chi', n_bins = 10)
    assert len(np.unique(res)) == 10

def test_merge_frame():
    res = merge(df, target = 'target', method = 'chi', n_bins = 10)
    assert len(np.unique(res['A'])) == 10



import os
import numpy as np
from setuptools import setup, find_packages, Extension


NAME = 'toad'


CURRENT_PATH = os.path.abspath(os.path.dirname(__file__))
VERSION_FILE = os.path.join(CURRENT_PATH, NAME, 'version.py')

def get_version():
    ns = {}
    with open(VERSION_FILE) as f:
        exec(f.read(), ns)
    return ns['__version__']


def get_ext_modules():
    from Cython.Build import cythonize

    extensions = [
        Extension('toad.c_utils', sources = ['toad/c_utils.pyx'], include_dirs = [np.get_include()]),
        Extension('toad.merge', sources = ['toad/merge.pyx'], include_dirs = [np.get_include()]),
    ]

    return cythonize(extensions)


def get_requirements(stage = None):
    file_name = 'requirements'

    if stage is not None:
        file_name = f"{file_name}-{stage}"
    
    requirements = []
    with open(f"{file_name}.txt", 'r') as f:
        for line in f:
            line = line.strip()
            if not line or line.startswith('-'):
                continue
            
            requirements.append(line)
    
    return requirements


setup(
    name = NAME,
    version = get_version(),
    description = 'Toad is dedicated to facilitating model development process, especially for a scorecard.',
    long_description = open('README.md', encoding = 'utf-8').read(),
    long_description_content_type = 'text/markdown',
    url = 'https://github.com/amphibian-dev/toad',
    author = 'ESC Team',
    author_email = 'secbone@gmail.com',
    packages = find_packages(exclude = ['tests']),
    include_dirs = [np.get_include()],
    ext_modules = get_ext_modules(),
    include_package_data = True,
    python_requires = '>=3.6',
    install_requires = get_requirements(),
    extras_require = {
        'nn': get_requirements('nn')
    },
    tests_require = get_requirements('test'),
    license = 'MIT',
    classifiers = [
        'Operating System :: POSIX',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: MacOS :: MacOS X',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    entry_points = {
        'console_scripts': [
            'toad = toad.cli:main',
        ],
    },
)



import pandas as pd
from pandas.api.types import is_numeric_dtype, is_object_dtype
from sklearn.ensemble import AdaBoostClassifier
from sklearn.metrics import make_scorer
from sklearn.model_selection import cross_val_score
import xlsxwriter
import toad
from toad.transform import WOETransformer





# 把index数字转换成区间
def rename_columns(index_list, bins, is_float=False):
    list = [int(x) for x in index_list]
    bins_list = []
    left = '[0,'
    if -9999999 in list:
        list.remove(-9999999)
        bins_list.append('[-9999999,-9999999]')
        left = '(-9999999,'
    for i in range(len(list)):
        if i == 0:
            bins_list.append(left + str(float('%.2f' % bins[list[0]]) if is_float else int(bins[list[0]])) + ')')
        elif i == len(list) - 1:
            bins_list.append('[' + str(float('%.2f' % bins[list[len(list) - 1] - 1]) if is_float else int(
                bins[list[len(list) - 1] - 1])) + ',inf)')
        else:
            bins_list.append(
                '[' + str(float('%.2f' % bins[list[i - 1]]) if is_float else int(bins[list[i - 1]])) + ',' + str(
                    float('%.2f' % bins[list[i]]) if is_float else int(bins[list[i]])) + ')')
    return bins_list


# 离散变量分组求坏账率
def self_bin_dispersed(X, Y):
    d1 = pd.DataFrame({"X": X, "Y": Y})
    d2 = d1.groupby("X")
    d3 = pd.DataFrame(columns=['bins', 'bad', 'fre', 'bad_rate'])
    for name, group in d2:
        bad_count = group.sum().Y
        total = group.count().Y
        bad_rate = group.mean().Y
        d3 = d3.append({'bins': name, 'bad': bad_count, 'fre': total, 'bad_rate': bad_rate}, ignore_index=True)
    d4 = d3.sort_values(by='bins')
    return d4


# 连续变量分组求坏账率
def self_bin_successive(X, Y, bins):
    d1 = pd.DataFrame({"X": X, "Y": Y, 'index': X})
    d2 = d1.groupby('index')
    d3 = pd.DataFrame()
    d3['index'] = d2.min().X
    d3['bad'] = d2.sum().Y
    d3['fre'] = d2.count().Y
    d3['bad_rate'] = d2.mean().Y
    d3.sort_values(by='index', inplace=True)
    index = d3['index'].tolist()
    bins_list = rename_columns(index, bins, True)
    d3['bins'] = bins_list
    d3.drop('index', axis=1, inplace=True)
    return d3


def draw_data(workbook,worksheet,data,**kwargs):
    start_index = kwargs['start_index']
    title = kwargs['title']
    columns_list =kwargs['columns']
    haed_format = workbook.add_format({'bold': True, 'valign': 'middle', 'align': 'center', 'border': 1})
    data_format = workbook.add_format({'valign': 'middle', 'align': 'center', 'border': 1})
    merge_format = workbook.add_format({'bold': 1, 'border': 1, 'align': 'center', 'valign': 'vcenter'})
    worksheet.merge_range(
        "A" + str(start_index + 1) + ":" + column_to_name(len(columns_list) - 1 ).upper() + str(start_index + 1),
        title, merge_format)
    worksheet.write_row('A' + str(start_index + 2), columns_list, haed_format)
    for j in range(len(columns_list)):
        worksheet.write_column(column_to_name(j) + str(start_index + 3), data[columns_list[j]], data_format)
    return len(data)+ 4 + start_index
# 单个变量的分组结果和图写入到excel
def save_bins_and_chart(worksheet, workbook, data, var_name, start_index):
    draw_data(workbook,worksheet,data,draw_index=False,start_index=start_index,title=var_name, columns = ['bins', 'bad', 'fre', 'bad_rate'])
    column_chart1 = workbook.add_chart({'type': 'column'})
    column_chart1.add_series({
        'name': '=分布!$B$' + str(start_index + 2),
        'categories': '=分布!$A$' + str(start_index + 3) + ':$A$' + str(start_index + 2 + len(data)),
        'values': '=分布!$B$' + str(start_index + 3) + ':$B$' + str(start_index + 2 + len(data)),

    })
    column_chart1.set_size({'width': 700, 'height': 380})
    column_chart1.add_series({
        'name': '=分布!$C$' + str(start_index + 2),
        'categories': '=分布!$A$' + str(start_index + 3) + ':$A$' + str(start_index + 2 + len(data)),
        'values': '=分布!$C$' + str(start_index + 3) + ':$C$' + str(start_index + 2 + len(data)),
    })
    line_chart1 = workbook.add_chart({'type': 'line'})
    line_chart1.add_series({
        'name': '=分布!$D$' + str(start_index + 2),
        'categories': '=分布!$A$' + str(start_index + 3) + ':$A$' + str(start_index + 2 + len(data)),
        'values': '=分布!$D$' + str(start_index + 3) + ':$D$' + str(start_index + 2 + len(data)),
        'y2_axis': True,
    })
    column_chart1.combine(line_chart1)
    column_chart1.set_title({'name': var_name})
    column_chart1.set_x_axis({'name': '分组'})
    column_chart1.set_y_axis({'name': '个数'})
    column_chart1.set_y2_axis({'name': '坏账率'})
    worksheet.insert_chart('F' + str(start_index + 1), column_chart1)
    return max(22 + start_index, start_index + len(data) + 4)


# 选择iv大于0.02或者iv前十的变量
def select_iv(quality,num,iv_threshold_value):
    if len(quality[quality['iv'] > 0]) < num:
        return quality[quality['iv'] > 0]
    else:
        quality.sort_values(by="iv", ascending=False)
        high_iv = quality[quality['iv'] > iv_threshold_value]
        return high_iv if len(high_iv) >= num else quality[0:num]


# 合并长尾数据并进行等步长分组
def merger_data(data, var, unique_num,is_merge_high=True):
    if is_numeric_dtype(data[var]) and data[var].nunique() > unique_num:
        data_miss = data[data[var] == -9999999]
        data_nomiss = data[data[var] != -9999999]
        merge_high_data = data_nomiss[var]
        if is_merge_high:
            merge_high_data = toad.utils.clip(data_nomiss[var], quantile=(None, .99))
        data_index, bins = toad.merge(merge_high_data, method='step', return_splits=True, n_bins=unique_num)
        temp = pd.DataFrame(data_index, columns=[var])
        temp = temp.append(data_miss[[var]], ignore_index=True)[var]
        target = data_nomiss.append(data_miss, ignore_index=True)['target']
        return temp, target, bins
    else:
        return data[var], None, None


# 两两指标做透视表
def crosstab_data(columns_var, row_var, data,unique_num,*args):
    columns_data, columns_target, columns_bins = merger_data(data, columns_var, unique_num,args[0])
    row_data, row_target, row_bins = merger_data(data, row_var, unique_num,args[1])
    result = pd.crosstab(row_data, columns_data, margins=True, dropna=False)
    if columns_bins is not None:
        columns = result.columns.tolist()
        columns.remove('All')
        columns_bins_list = rename_columns(columns, columns_bins, args[2])
        columns_bins_list.append('All')
        result.set_axis(columns_bins_list, axis=1, inplace=True)
    if row_bins is not None:
        index = result.index.tolist()
        index.remove('All')
        index_bins_list = rename_columns(index, row_bins, args[3])
        index_bins_list.append('All')
        result.set_axis(index_bins_list, axis=0, inplace=True)
    return result


# 写入所有高iv的变量分组和图到excel
def all_var_pic(high_iv_var_list, data, workbook, worksheet,start_index,unique_num):
    for i in range(len(high_iv_var_list)):
        var_name = high_iv_var_list[i]
        result = None
        data_index, target, bins = merger_data(data, var_name, unique_num,True)
        if bins is not None:
            result = self_bin_successive(data_index, target, bins)
        else:
            result = self_bin_dispersed(data[var_name], data['target'])
        start_index = save_bins_and_chart(worksheet, workbook, result, var_name, start_index)
    return start_index


# 写入所有指标的逾期天数透视表到excel
def save_overdue_crosstab(data, var_list, workbook, worksheet,start_index,unique_num):
    for i in range(len(var_list)):
        var = var_list[i]
        result = crosstab_data(var, 'overdue_days', data,unique_num,True, False, True, False)
        columns = ['overdue_days'] + result.columns.tolist()
        result['overdue_days'] = result.index
        start_index = draw_data(workbook,worksheet,result,start_index=start_index,title=var,columns=columns)
    return start_index


# 进入模型的变量分组，少于10个的一个一组，大于10个的,按照10%分为一组
def var_bins(quality):
    quality.sort_values(by='iv', ascending=False, inplace=True)
    var_group_list = []
    if len(quality) < 10:
        for temp in quality.index.tolist():
            var_group_list.append([temp])
    else:
        bins = pd.qcut(range(len(quality)), 10, labels=False)
        df_var = pd.DataFrame(columns=['num', 'var', 'iv'])
        df_var['num'] = bins
        df_var['var'] = quality.index
        for group, temp in df_var.groupby(by='num'):
            var_group_list.append(temp['var'].tolist())
    return var_group_list


# 用woe替换离散变量
def replace_with_woe(train_data, test_data, exclude_var, target='target'):
    all_var = train_data.columns.tolist()
    for var in all_var:
        if var not in exclude_var and is_object_dtype(train_data[var]):
            woe = WOETransformer().fit(train_data[var], train_data[target])
            train_data[var] = woe.transform(train_data[var])
            train_data[var].astype('float64')
            test_data[var]  = woe.transform(test_data[var])
            test_data[var].astype('float64')
    return train_data, test_data


# 按照变量分组循环建模，并计算ks
def compute_ks(data_train, data_test, quality, self_var_list, self_var_ks, result, target='target'):
    test_var_list = var_bins(quality)
    best_ks = self_var_ks
    best_input_var = []
    for temp in test_var_list:
        input_var = best_input_var + temp
        model, param = param_opt(data_train[input_var + self_var_list], data_train[target])
        train_ks = get_ks(model, data_train[input_var + self_var_list], data_train[target])
        test_ks = get_ks(model, data_test[input_var + self_var_list], data_test[target])
        if best_ks < test_ks:
            best_ks = test_ks
            best_input_var = best_input_var + input_var
        result = result.append({'train_ks': train_ks, 'test_ks': test_ks, 'var': str(input_var)}, ignore_index=True)
    return result


# 把序列号转成excel列名
def column_to_name(index):
    if index > 25:
        ch1 = chr(index % 26 + 65)
        ch2 = chr(int(index / 26) + 64)
        return ch2 + ch1
    else:
        return chr(index % 26 + 65)


def ks_score(y, y_pred):
    return toad.KS( y_pred[:,1],y)

scorer = make_scorer(ks_score, needs_proba = True)

def test_score(p, x_train, y_train):
    model_ada = AdaBoostClassifier(
        algorithm='SAMME.R',
        base_estimator=None,
        learning_rate=0.1,
        n_estimators=p,
        random_state=123,
    )
    score_mean = cross_val_score(model_ada, x_train, y_train, cv=10, scoring = scorer).mean()
    return score_mean

def findMax(start, end, results, train_x, train_y):
    if end - start <= 10:
        if results[start] > results[end]:
            return start, results[start]
        else:
            return end, results[end]
    need_test = int((start + end) / 2)
    results.at[need_test] = test_score(need_test, train_x, train_y)
    if results[need_test] > max(results[start], results[end]) or results[need_test] < min(results[start], results[end]):
        p1, r1 = findMax(start, need_test, results, train_x, train_y)
        p2, r2 = findMax(need_test, end, results, train_x, train_y)
        if r1 > r2:
            return p1, r1
        else:
            return p2, r2
    elif results[start] > results[end]:
        return findMax(start, need_test, results, train_x, train_y)
    else:
        return findMax(need_test, end, results, train_x, train_y)
def param_opt(train_x, train_y, start = 100, end = 700):
    results = pd.Series()
    results.at[start] = test_score(start, train_x, train_y)
    results.at[end] = test_score(end, train_x, train_y)
    param, mean = findMax(start, end, results, train_x, train_y)
    model = AdaBoostClassifier(
        algorithm='SAMME.R',
        learning_rate=0.1,
        n_estimators=param,
        random_state=123,
    )
    model.fit(train_x, train_y)
    return model,param

def get_ks(model, X, Y):
    pre = model.predict_proba(X)
    return ks_score(Y, pre)



def evaluate(test_data, excel_name = 'report.xlsx', num = 10, iv_threshold_value = 0.02, unique_num = 20, overdue_days = False, self_data = None):
    # 测试数据iv等信息并写入excel
    workbook = xlsxwriter.Workbook(excel_name)
    quality = toad.quality(test_data.drop(columns=["loan_apply_no"]), target="target")
    quality.sort_values(by='iv',ascending=False,inplace=True)
    quality['var_name'] = quality.index
    draw_data(workbook, workbook.add_worksheet('quality'), quality, start_index=0, title='变量探查结果',
              columns=['var_name', 'iv', 'gini', 'entropy', 'unique'])
    print("quality 计算完毕")
    # 选择高iv的测试变量分组并写入excel
    quality = quality.replace("--", -1)
    high_iv_var = select_iv(quality,num,iv_threshold_value)
    high_iv_var_list = high_iv_var.index.tolist()
    all_var_pic(high_iv_var_list, test_data, workbook, workbook.add_worksheet("分布"), 0,unique_num)
    print("变量分组处理完毕")
    all_data = None
    if overdue_days or self_data is not None:
        all_data = pd.merge(self_data, test_data.drop(columns=['target']), how='left', on='loan_apply_no')
    if overdue_days:
        # 数据合并
        test_var = test_data.columns.tolist()
        test_var.remove("loan_apply_no")
        test_var.remove("target")
        save_overdue_crosstab(all_data, test_var, workbook, workbook.add_worksheet("逾期天数"), 0,unique_num=unique_num)
        print("逾期天数处理完毕")
    if self_data is not None:
        # train和test数据集划分
        data_train = all_data[all_data['SMP_N'] == 'dev_smp']
        data_test = all_data[all_data['SMP_N'] == 'ver_smp']
        # 离散变量做woe处理
        data_train, data_test = replace_with_woe(data_train, data_test, exclude_var=['SMP_N', 'loan_apply_no'],
                                                 target='target')
        # 循环加入变量建模并计算模型的ks最后写入excel
        self_data_var_list = self_data.columns.tolist()
        drop_var = ['loan_apply_no', 'SMP_N', 'target']
        for var in drop_var:
            self_data_var_list.remove(var)
        if 'overdue_days' in self_data_var_list:
            self_data_var_list.remove('overdue_days')
        model, param = param_opt(data_train[self_data_var_list], data_train['target'])
        train_ks = get_ks(model, data_train[self_data_var_list], data_train['target'])
        test_ks = get_ks(model, data_test[self_data_var_list], data_test['target'])
        result = pd.DataFrame(columns=['train_ks', 'test_ks', 'var'])
        result = result.append({'train_ks': train_ks, 'test_ks': test_ks, 'var': '[]'}, ignore_index=True)
        result = compute_ks(data_train, data_test, quality, self_data_var_list, test_ks, result, target='target')
        draw_data(workbook, workbook.add_worksheet("ks"), result, start_index=0, title='ks测试结果',
                  columns=result.columns.tolist())
        print("模型计算完毕")
    workbook.close()
